
from pathlib import Path

from pytest import mark

import asreview as asr
from asreview.readers import ASReviewData


@mark.parametrize(
    "test_file,n_lines,labels,ignore_col",
    [
        ("embase.csv", 6, None, []),
        ("embase.ris", 6, None, []),
        ("generic.csv", 2, None, []),
        ("generic_labels.csv", 6, [1, 0, 1, 0, 1, 0], []),
        ("generic.ris", 2, None, []),
        ("generic_labels.ris", 2, [1, 0], []),
        ("pubmed_zotero.ris", 6, None, []),
        ("pubmed_endnote.txt", 6, None, []),
        ("scopus.ris", 6, None, []),
        ("ovid_zotero.ris", 6, None, []),
        ("proquest.ris", 6, None, []),
        ("pubmed.xml", 10, None, []),
    ])
def test_reader(test_file, n_lines, labels, ignore_col):
    fp = Path("test", "demo_data", test_file)
    as_data = asr.ASReviewData.from_file(fp)
    assert as_data.raw_df.shape[0] == n_lines

    cols = ['title', 'abstract', 'authors', 'article_id']
    cols = [col for col in cols if col not in ignore_col]
    if labels is not None:
        cols.append('labels')
        for i in range(n_lines):
            assert as_data.labels[i] == labels[i]

    for col in cols:
        print(col)
        assert len(getattr(as_data, col)) == n_lines


def test_csv_write_data():
    fp_in = Path("test", "demo_data", "generic_labels.csv")
    fp_out = Path("test", "out_data", "generic_out.csv")
    asr_data = ASReviewData.from_file(fp_in)
    asr_data.to_csv(fp_out, labels=[0, 1, 0, 1, 0, 1])



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command Line Interface (CLI) for ASReview project."""
import argparse
import logging
import sys
import warnings
from argparse import RawTextHelpFormatter

from asreview.ascii import welcome_message


import numpy as np  # noqa
import tensorflow as tf # noqa
try:
    tf.logging.set_verbosity(tf.logging.ERROR)
except AttributeError:
    logging.getLogger("tensorflow").setLevel(logging.ERROR)

from asreview import __version__  # noqa
from asreview.review import review_oracle, review_simulate  # noqa
from asreview.config import AVAILABLE_CLI_MODI  # noqa
from asreview.config import (  # noqa
    DEFAULT_MODEL, DEFAULT_QUERY_STRATEGY, DEFAULT_BALANCE_STRATEGY,
    DEFAULT_N_INSTANCES, DEFAULT_N_PRIOR_INCLUDED, DEFAULT_N_PRIOR_EXCLUDED)


# Descriptions

PROG_DESCRIPTION = f"""
Automated Systematic Review (ASReview).

Use one of the modi: '{AVAILABLE_CLI_MODI[0]}' or '{AVAILABLE_CLI_MODI[1]}'
"""

PROG_DESCRIPTION_SIMULATE = """
Automated Systematic Review (ASReview) for simulation runs.

The simulation modus is used to measure the performance of our
software on existing systematic reviews. The software shows how many
papers you could have potentially skipped during the systematic
review."""

PROG_DESCRIPTION_ORACLE = """
Automated Systematic Review (ASReview) with interaction with oracle.

The oracle modus is used to perform a systematic review with
interaction by the reviewer (the ‘oracle’ in literature on active
learning). The software presents papers to the reviewer, whereafter
the reviewer classifies them."""


def _parse_arguments(mode, prog=sys.argv[0]):
    """Argument parser for oracle and simulate.

    Parameters
    ----------
    mode : str
        The mode to run ASReview. Options:
        'simulate' and 'oracle'.
    prog : str
        The program name. For example 'asreview'.

    Returns
    -------
    argparse.ArgumentParser
        Configured argparser.
    """

    if mode == "simulate":
        prog_description = PROG_DESCRIPTION_SIMULATE
    elif mode == "oracle":
        prog_description = PROG_DESCRIPTION_ORACLE
    else:
        prog_description = ""

    # parse arguments if available
    parser = argparse.ArgumentParser(
        prog=prog,
        description=prog_description,
        formatter_class=RawTextHelpFormatter
    )
    # File path to the data.
    parser.add_argument(
        "dataset",
        type=str,
        metavar="X",
        help="File path to the dataset or one of the built-in datasets."
    )
    # Active learning parameters
    parser.add_argument(
        "-m", "--model",
        type=str,
        default=DEFAULT_MODEL,
        help=f"The prediction model for Active Learning. "
             f"Default '{DEFAULT_MODEL}'.")  #noqa
    parser.add_argument(
        "-q", "--query_strategy",
        type=str,
        default=DEFAULT_QUERY_STRATEGY,
        help=f"The query strategy for Active Learning. "
             f"Default '{DEFAULT_QUERY_STRATEGY}'.")  #noqa
    parser.add_argument(
        "-b", "--balance_strategy",
        type=str,
        default=DEFAULT_BALANCE_STRATEGY,
        help="Data rebalancing strategy mainly for RNN methods. Helps against"
             " imbalanced dataset with few inclusions and many exclusions. "
             f"Default '{DEFAULT_BALANCE_STRATEGY}'")
    parser.add_argument(
        "--n_instances",
        default=DEFAULT_N_INSTANCES,
        type=int,
        help="Number of papers queried each query."
             f"Default {DEFAULT_N_INSTANCES}.")
    parser.add_argument(
        "--n_queries",
        type=int,
        default=None,
        help="The number of queries. By default, the program "
             "stops after all documents are reviewed or is "
             "interrupted by the user."
    )
    parser.add_argument(
        "-n", "--n_papers",
        type=int,
        default=None,
        help="The number of papers to be reviewed. By default, "
             "the program stops after all documents are reviewed or is "
             "interrupted by the user."
    )
    parser.add_argument(
        "--embedding",
        type=str,
        default=None,
        dest='embedding_fp',
        help="File path of embedding matrix. Required for LSTM models."
    )
    # Configuration file with model/balance/query parameters.
    parser.add_argument(
        "--config_file",
        type=str,
        default=None,
        help="Configuration file with model parameters"
    )
    # Initial data (prior knowledge)
    parser.add_argument(
        "--prior_included",
        default=None,
        type=int,
        nargs="*",
        help="A list of included papers.")

    parser.add_argument(
        "--prior_excluded",
        default=None,
        type=int,
        nargs="*",
        help="A list of excluded papers. Optional.")

    parser.add_argument(
        "--extra_dataset",
        default=[],
        action='append',
        help="A dataset with labels to improve training. Can be used multiple"
             " times."
    )

    # these flag are only available for the simulation modus
    if mode == "simulate":

        # Initial data (prior knowledge)
        parser.add_argument(
            "--n_prior_included",
            default=DEFAULT_N_PRIOR_INCLUDED,
            type=int,
            help="Sample n prior included papers. "
                 "Only used when --prior_included is not given. "
                 f"Default {DEFAULT_N_PRIOR_INCLUDED}")

        parser.add_argument(
            "--n_prior_excluded",
            default=DEFAULT_N_PRIOR_EXCLUDED,
            type=int,
            help="Sample n prior excluded papers. "
                 "Only used when --prior_excluded is not given. "
                 f"Default {DEFAULT_N_PRIOR_EXCLUDED}")

    # logging and verbosity
    parser.add_argument(
        "--log_file", "-l",
        default=None,
        type=str,
        help="Location to store the log results."
    )
    parser.add_argument(
        "--save_model",
        default=None,
        type=str,
        dest='save_model_fp',
        help="Location to store the model and weights. "
             "Only works for Keras/RNN models. "
             "End file extension with '.json'."
    )
    parser.add_argument(
        "--verbose", "-v",
        default=0,
        type=int,
        help="Verbosity")

    return parser


def _review_general(mode="oracle"):
    parser = _parse_arguments(mode, prog="asreview " + mode)
    args = parser.parse_args(sys.argv[2:])

    args_dict = vars(args)
    path = args_dict.pop("dataset")

    verbose = args_dict.get("verbose", 0)
    if verbose == 0:
        logging.getLogger().setLevel(logging.WARNING)
    elif verbose == 1:
        logging.getLogger().setLevel(logging.INFO)
    elif verbose >= 2:
        logging.getLogger().setLevel(logging.DEBUG)

    print(welcome_message(mode))
    if mode == "oracle":
        review_oracle(path, **args_dict)
    elif mode == "simulate":
        review_simulate(path, **args_dict)


def main_depr():
    warnings.warn("'asr' has been renamed to "
                  "'asreview', it will be removed in the future.\n",
                  np.VisibleDeprecationWarning)
    main()


def main():
    # launch asr interactively
    if len(sys.argv) > 1 and sys.argv[1] in ["oracle", "simulate"]:
        _review_general(sys.argv[1])

    # no valid sub command
    else:
        parser = argparse.ArgumentParser(
            prog="asreview",
            description=PROG_DESCRIPTION
        )
        parser.add_argument(
            "subcommand",
            nargs="?",
            type=lambda x: isinstance(x, str) and x in AVAILABLE_CLI_MODI,
            default=None,
            help=f"The subcommand to launch. Available commands: "
            f"{AVAILABLE_CLI_MODI}"
        )

        # version
        parser.add_argument(
            "-V", "--version",
            action='store_true',
            help="print the ASR version number and exit")

        args, _ = parser.parse_known_args()

        # output the version
        if args.version:
            print(__version__)
            return

        parser.print_help()


# execute main function
if __name__ == "__main__":
    main()



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Random sampling strategy."""

import numpy as np

from asreview.query_strategies.base import NotProbaQueryStrategy


class RandomQuery(NotProbaQueryStrategy):
    "Random sampling query strategy."
    name = "random"

    def _query(self, X, pool_idx, n_instances=1):
        n_samples = len(pool_idx)
        query_idx = np.random.choice(
            np.arange(n_samples),
            n_instances,
            replace=False
        )
        return pool_idx[query_idx], X[pool_idx[query_idx]]



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from tensorflow.keras.constraints import MaxNorm
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Embedding
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import LSTM
from tensorflow.keras.layers import MaxPooling1D
from tensorflow.keras.models import Sequential
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier

from asreview.models.lstm_base import _get_optimizer
from asreview.models.base import BaseTrainModel
from asreview.utils import _set_class_weight


class LSTMPoolModel(BaseTrainModel):
    """ LSTM pool class.

    LSTM model consisting of an embedding layer, one LSTM layer, and one
    max pooling layer.
    """
    name = "lstm-pool"

    def __init__(self, embedding_matrix=None, backwards=True, dropout=0.4,
                 optimizer="rmsprop", lstm_out_width=20, lstm_pool_size=128,
                 learn_rate=1.0, verbose=0, batch_size=32, epochs=35,
                 shuffle=False, class_weight=30.0):
        """Initialize the LSTM pool model.

        Arguments
        ---------
        embedding_matrix: np.array
            Embedding matrix to use with LSTM model.
        backwards: bool
            Whether to have a forward or backward LSTM.
        dropout: float
            Value in [0, 1.0) that gives the dropout and recurrent
            dropout rate for the LSTM model.
        optimizer: str
            Optimizer to use.
        lstm_out_width: int
            Output width of the LSTM.
        lstm_pool_size: int
            Size of the pool, must be a divisor of max_sequence_length.
        learn_rate: float
            Learn rate multiplier of default learning rate.
        verbose: int
            Verbosity.
        batch_size: int
            Size of the batch size for the LSTM model.
        epochs: int
            Number of epochs to train the LSTM model.
        shuffle: bool
            Whether to shuffle the data before starting to train.
        class_weight: float
            Class weight for the included papers.
        """
        super(LSTMPoolModel, self).__init__()
        self.embedding_matrix = embedding_matrix
        self.backwards = backwards
        self.dropout = dropout
        self.optimizer = optimizer
        self.lstm_out_width = lstm_out_width
        self.learn_rate = learn_rate
        self.verbose = verbose
        self.batch_size = batch_size
        self.epochs = epochs
        self.shuffle = shuffle
        self.class_weight = _set_class_weight(class_weight)
        self.lstm_pool_size = lstm_pool_size
        self._model = None
        self.sequence_length = None

    def fit(self, X, y):
        sequence_length = X.shape[1]
        if self._model is None or sequence_length != self.sequence_length:
            self.sequence_length = sequence_length
            keras_model = _create_lstm_pool_model(
                embedding_matrix=self.embedding_matrix,
                backwards=self.backwards,
                dropout=self.dropout, optimizer=self.optimizer,
                max_sequence_length=sequence_length,
                lstm_out_width=self.lstm_out_width,
                learn_rate=self.learn_rate, verbose=self.verbose)
            self._model = KerasClassifier(keras_model, verbose=self.verbose)
        self._model.fit(X, y, batch_size=self.batch_size, epochs=self.epochs,
                        shuffle=self.shuffle, class_weight=self.class_weight,
                        verbose=self.verbose)

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_choices = {}
        hyper_space = {
            "mdl_dropout": hp.uniform("mdl_dropout", 0, 0.9),
            "mdl_lstm_out_width": hp.quniform("mdl_lstm_out_width", 1, 50, 1),
            "mdl_dense_width": hp.quniform("mdl_dense_width", 1, 200, 1),
            "mdl_learn_rate_mult": hp.lognormal("mdl_learn_rate_mult", 0, 1)
        }
        return hyper_space, hyper_choices


def _create_lstm_pool_model(embedding_matrix,
                            backwards=True,
                            dropout=0.4,
                            optimizer='rmsprop',
                            max_sequence_length=1000,
                            lstm_out_width=20,
                            lstm_pool_size=100,
                            learn_rate=1.0,
                            verbose=1):
    """Return callable lstm model.
    Returns
    -------
    callable:
        A function that return the Keras Sklearn model when
        called.

    """
    # The Sklearn API requires a callable as result.
    # https://keras.io/scikit-learn-api/
    def model_wrapper():
        model = Sequential()

        # add first embedding layer with pretrained wikipedia weights
        model.add(
            Embedding(
                embedding_matrix.shape[0],
                embedding_matrix.shape[1],
                weights=[embedding_matrix],
                input_length=max_sequence_length,
                trainable=False
            )
        )

        # add LSTM layer
        model.add(
            LSTM(
                lstm_out_width,
                input_shape=(max_sequence_length, ),
                go_backwards=backwards,
                dropout=dropout,
                recurrent_dropout=dropout,
                return_sequences=True,
                kernel_constraint=MaxNorm(),
            )
        )

        model.add(
            MaxPooling1D(
                pool_size=lstm_pool_size,
            )
        )
        model.add(
            Flatten()
        )

        # Add output layer
        model.add(
            Dense(
                1,
                activation='sigmoid'
            )
        )

        optimizer_fn = _get_optimizer(optimizer, learn_rate)

        # Compile model
        model.compile(
            loss='binary_crossentropy', optimizer=optimizer_fn,
            metrics=['acc'])

        if verbose >= 1:
            model.summary()

        return model

    return model_wrapper



import string
import random
import os
import urllib.request

from asreview.feature_extraction.embedding_lstm import load_embedding
from asreview.feature_extraction.embedding_lstm import sample_embedding


def random_words(n_words=1000, other_word_dict={}):
    """ Generator of random (ascii) words.

    Parameters
    ----------
    n_words: int
        Number of words to generate

    Returns
    -------

    str:
        List of words
    """

    word_dict = {}
    i = 0
    # Generate words until we have enough.
    while len(word_dict) < n_words:
        n_letters = random.randint(1, 10)
        new_word = ""
        for _ in range(n_letters):
            new_word += random.choice(string.ascii_letters)
        if new_word not in word_dict and new_word not in other_word_dict:
            word_dict[new_word] = i
            i += 1
    return list(word_dict)


def random_sample_embedding(words, n_samples=200):
    """ Generator of sampled embeddings.

    Parameters
    ----------
    words: str
        List of words to sample from; len(words) >= n_samples.
    n_samples: int
        Number of samples to take

    Returns
    -------
    dict:
        Dictionary containing sampled words as keys and order as values.
    """

    assert len(words) >= n_samples
    keys = random.sample(words, n_samples)
    order = random.sample(range(1, n_samples+1), n_samples)
    for i in range(1, n_samples+1):
        assert i in order
    word_index = dict(zip(keys, order))
    assert len(keys) == n_samples
    assert len(order) == n_samples
    print(keys)
    print(order)
    print(words)
    assert len(word_index) == n_samples
    return word_index


def random_embedding(words, emb_vec_dim=300):
    """ Generator of random embedding.

    Parameters
    ----------
    words: str
        List of words to generate embedding for.
    emb_vec_dim: int
       Dimension of embedding vectors.

    Returns
    -------
    dict:
        Random embedding dictionary for all input words.
    """
    full_embedding = {}
    for word in words:
        new_rand = [random.random() for _ in range(emb_vec_dim)]
        full_embedding[word] = new_rand
    return full_embedding


def write_random_embedding(full_embedding, tmpfile):
    """ Write an embedding to a file.

    Parameters
    ----------
    tmpfile: str
        Temporary file created by pytest.
    full_embedding: dict
        Embedding without sampling.
    """

    n_words = len(full_embedding)
    emb_vec_dim = len(full_embedding[list(full_embedding)[0]])
    with open(tmpfile, "w") as f:
        f.write(f"{n_words} {emb_vec_dim}\n")
        for word in list(full_embedding):
            f.write(word+" ")
            new_rand_str = [str(x) for x in full_embedding[word]]
            # Add extra space at the end of the line for compatibility.
            f.write(" ".join(new_rand_str)+" \n")


def check_embedding(emb, full_embedding, n_samples, emb_vec_dim):
    """ Do the checking for the loaded embedding.

    Parameters
    ----------
    emb: dict
        Embedding read from a file, potentially subsampled.
    full_embedding: dict
        Embedding as created in the unit test directly.
    n_samples: int
        Number of samples taken, if not sampled n_samples == n_words.
    emb_vec_dim: int
        Dimension of the embedding vectors.
    """
    assert len(emb) == n_samples
    for key in emb:
        assert emb[key].size == emb_vec_dim
        for i in range(emb[key].size):
            assert abs(emb[key][i] - full_embedding[key][i]) < 1e-5
    assert isinstance(emb, (dict,))


def test_load_embedding(tmpdir):
    """ Unit test for load_embedding function. """
    n_words = 100
    n_samples = 30
    emb_vec_dim = 133

    # Generate embedding.
    words = random_words(n_words)
    word_index = random_sample_embedding(words, n_samples)
    tmpfile = os.path.join(tmpdir, "emb.dat")
    full_embedding = random_embedding(words, emb_vec_dim)

    write_random_embedding(full_embedding, tmpfile)

    # Test with one worker process, no sampling.
    emb1 = load_embedding(tmpfile, word_index=None, n_jobs=1)
    check_embedding(emb1, full_embedding, n_words, emb_vec_dim)

    # Test with all cores, sampling.
    emb2 = load_embedding(tmpfile, word_index, n_jobs=-1)
    check_embedding(emb2, full_embedding, n_samples, emb_vec_dim)

    # Test with 3+1 cores, sampling.
    emb3 = load_embedding(tmpfile, word_index, n_jobs=3)
    check_embedding(emb3, full_embedding, n_samples, emb_vec_dim)


def test_sample_embedding():
    """ Unit test for sample_embedding """
    n_words = 50
    n_samples = 55
    emb_vec_dim = 133
    emb_extra = 10

    # Generate embedding.
    words = random_words(n_words)
    all_words = random_words(emb_extra, words) + words
    word_index = random_sample_embedding(all_words, n_samples)
    full_embedding = random_embedding(words, emb_vec_dim)
    emb_matrix = sample_embedding(full_embedding, word_index)

    assert emb_matrix.shape == (n_samples+1, emb_vec_dim)
    for key in word_index:
        i_row = word_index[key]
        if key in words:
            for i in range(emb_vec_dim):
                assert emb_matrix[i_row][i] == full_embedding[key][i]
            for i in range(emb_vec_dim, emb_vec_dim):
                assert emb_matrix[i_row][i] == 0


def test_embedding_link():
    """ Test if wikipedia embedding still exists. """
    url = "https://dl.fbaipublicfiles.com" \
          "/fasttext/vectors-crawl/cc.en.300.vec.gz"
    assert urllib.request.urlopen(url).getcode() == 200



import os
from pathlib import Path

import numpy as np

from asreview.logging import JSONLogger, HDF5Logger, DictLogger
from asreview import open_logger
from asreview.settings import ASReviewSettings


def test_read_json_logger():

    log_fp = Path("test", "log_files", "test_1_inst.json")

    with open_logger(str(log_fp)) as logger:
        assert isinstance(logger, JSONLogger)


def test_read_hdf5_logger():
    log_fp = Path("test", "log_files", "test_1_inst.h5")
    with open_logger(str(log_fp)) as logger:
        assert isinstance(logger, HDF5Logger)


def test_read_dict_logger():
    with open_logger(None) as logger:
        assert isinstance(logger, DictLogger)


def test_write_json_logger(tmpdir):
    check_write_logger(tmpdir, 'test.json')


def test_write_hdf5_logger(tmpdir):
    check_write_logger(tmpdir, 'test.h5')


def test_write_dict_logger(tmpdir):
    check_write_logger(tmpdir, None)


def check_write_logger(tmpdir, log_file):
    if log_file is not None:
        log_fp = os.path.join(tmpdir, log_file)
    else:
        log_fp = None

    settings = ASReviewSettings(mode="simulate", model="nb",
                                query_strategy="rand_max",
                                balance_strategy="simple",
                                feature_extraction="tfidf")

    n_records = 6
    n_half = int(n_records/2)
    start_labels = np.full(n_records, np.nan, dtype=np.int)
    labels = np.zeros(n_records, dtype=np.int)
    labels[::2] = np.ones(n_half, dtype=np.int)
    methods = np.full((n_records), "initial")
    methods[2::] = np.full((int(n_records-2)), "random")
    methods[2::2] = np.full((int((n_records-2)/2)), "max")

    with open_logger(log_fp) as logger:
        logger.add_settings(settings)
        logger.set_labels(start_labels)
        current_labels = np.copy(start_labels)
        for i in range(n_records):
            query_i = int(i/2)
            proba = None
            if i >= 2 and (i % 2) == 0:
                proba = np.random.rand(n_records)
            logger.add_classification([i], [labels[i]], [methods[i]], query_i)
            if proba is not None:
                logger.add_proba(np.arange(i+1, n_records), np.arange(i+1),
                                 proba, query_i)
            current_labels[i] = labels[i]
            logger.set_labels(current_labels)
            check_logger(logger, i, query_i, labels, methods, proba)


def check_logger(logger, label_i, query_i, labels, methods, proba):
    n_records = len(labels)

    log_labels = logger.get("labels")
    assert len(log_labels) == len(labels)
    for i in range(label_i+1):
        assert log_labels[i] == labels[i]
    for i in range(label_i+1, n_records):
        assert log_labels[i] == np.full(1, np.nan, dtype=np.int)[0]

    result_dict = logger.to_dict()
    cur_i = 0
    for qi in range(query_i+1):
        res = result_dict["results"][qi]
        max_i = cur_i + len(res["label_methods"])
        for j in range(len(res["label_methods"])):
            assert res["label_methods"][j] == methods[j+cur_i]
        assert np.all(res["label_idx"] == np.arange(cur_i, max_i))
        assert np.all(res["inclusions"] == labels[cur_i:max_i])
        cur_i = max_i

    if proba is not None:
        res = result_dict["results"][query_i]
        assert np.all(res["proba"] == proba)
        assert np.all(res["pool_idx"] == list(range(label_i+1, n_records)))
        assert np.all(res["train_idx"] == list(range(0, label_i+1)))



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import log, floor

import numpy as np

from asreview.balance_strategies.base import BaseBalance
from asreview.balance_strategies.simple import SimpleBalance


class DoubleBalance(BaseBalance):
    """
    Class to get the two way rebalancing function and arguments.
    It super samples ones depending on the number of 0's and total number
    of samples in the training data.
    """
    name = "double-balance"

    def __init__(self, a=2.155, alpha=0.94, b=0.789, beta=1.0):
        """Initialize the double balance strategy.

        Arguments
        ---------
        a: float
            Governs the weight of the 1's. Higher values mean linearly more 1's
            in your training sample.
        alpha: float
            Governs the scaling the weight of the 1's, as a function of the
            ratio of ones to zeros. A positive value means that the lower the
            ratio of zeros to ones, the higher the weight of the ones.
        b: float
            Governs how strongly we want to sample depending on the total
            number of samples. A value of 1 means no dependence on the total
            number of samples, while lower values mean increasingly stronger
            dependence on the number of samples.
        beta: float
            Governs the scaling of the weight of the zeros depending on the
            number of samples. Higher values means that larger samples are more
            strongly penalizing zeros.
        """
        super(DoubleBalance, self).__init__()
        self.a = a
        self.alpha = alpha
        self.b = b
        self.beta = beta
        self.fallback_model = SimpleBalance()

    def sample(self, X, y, train_idx, shared):
        one_idx = train_idx[np.where(y[train_idx] == 1)]
        zero_idx = train_idx[np.where(y[train_idx] == 0)]

        # Fall back to simple sampling if we have only ones or zeros.
        if len(one_idx) == 0 or len(zero_idx) == 0:
            self.fallback_model.sample(X, y, train_idx, shared)

        n_one = len(one_idx)
        n_zero = len(zero_idx)
        n_train = n_one + n_zero

        # Compute the weights.
        one_weight = _one_weight(n_one, n_zero, self.a, self.alpha)
        zero_weight = _zero_weight(n_one+n_zero, self.b, self.beta)
        tot_zo_weight = one_weight * n_one + zero_weight * n_zero
        n_one_train = random_round(one_weight*n_one*n_train/tot_zo_weight)
        n_one_train = max(1, min(n_train-2, n_one_train))
        n_zero_train = n_train-n_one_train

        # Get random ones and zeros.
        one_train_idx = fill_training(one_idx, n_one_train)
        zero_train_idx = fill_training(zero_idx, n_zero_train)

        all_idx = np.concatenate([one_train_idx, zero_train_idx])
        np.random.shuffle(all_idx)

        return X[all_idx], y[all_idx]

    def full_hyper_space(self):
        from hyperopt import hp
        parameter_space = {
            "bal_a": hp.lognormal("bal_a", 0, 1),
            "bal_alpha": hp.uniform("bal_alpha", 0, 2),
            "bal_b": hp.uniform("bal_b", 0, 1),
            # "bal_beta": hp.uniform("bal_beta", 0, 2),
        }
        return parameter_space, {}


def _one_weight(n_one, n_zero, a, alpha):
    """Get the weight of the ones."""
    weight = a * (n_one/n_zero)**(-alpha)
    return weight


def _zero_weight(n_read, b, beta):
    """Get the weight of the zeros."""
    weight = 1 - (1-b) * (1+log(n_read))**(-beta)
    return weight


def random_round(value):
    """Round up or down, depending on how far the value is.

    For example: 8.1 would be rounded to 8, 90% of the time, and rounded
    to 9, 10% of the time.
    """
    base = int(floor(value))
    if np.random.rand() < value-base:
        base += 1
    return base


def fill_training(src_idx, n_train):
    """Copy/sample until there are n_train indices sampled.
    """
    # Copy as many as we need, rounded down.
    n_copy = np.int(n_train/len(src_idx))
    n_sample = n_train - n_copy*len(src_idx)
    dest_idx = np.tile(src_idx, n_copy).reshape(-1)

    # Add samples to finish up.
    dest_idx = np.append(dest_idx,
                         np.random.choice(src_idx, n_sample, replace=False))
    return dest_idx



#!/usr/bin/env python

# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import sqrt

import numpy as np


def simulate_score(one_dict, all_dict, n_run=10000):
    total_one = np.sum([x for x in one_dict.values()])
    total = np.sum([x for x in all_dict.values()])
    sim_scores = []
    for _ in range(n_run):
        one_idx = np.random.choice(range(total), total_one, replace=False)
        one_idx = np.sort(one_idx)
        new_one_dict = {}
        cur_all_idx = 0
        cur_one_idx = 0
        for key in all_dict:
            cur_all_idx += all_dict[key]
            while cur_one_idx < len(one_idx) and one_idx[cur_one_idx] < cur_all_idx:
                if key in new_one_dict:
                    new_one_dict[key] += 1
                else:
                    new_one_dict[key] = 1
                cur_one_idx += 1
        try:
            sim_scores.append(cluster_score(new_one_dict, all_dict))
        except ZeroDivisionError:
            print(new_one_dict, all_dict)
            raise

    return np.average(sim_scores), np.std(sim_scores)


def cluster_score(one_dict, all_dict):
    tp = 0
    fn = 0
    fp = 0
    total = np.sum(list(one_dict.values()))
    for key, n_total in all_dict.items():
        n_one = one_dict.get(key, 0)
        n_zero = n_total-n_one
        tp += n_one*(n_one - 1)/2
        fn += n_zero*n_one
        fp += n_one*(total-n_one)
    return tp/sqrt(1+(tp+fn)*(tp+fp))


def normalized_cluster_score(prediction, labels):
    one_dict, all_dict = get_one_all_dict(prediction, labels)
    score = cluster_score(one_dict, all_dict)
    avg, sigma = simulate_score(one_dict, all_dict)
    return (score-avg)/sigma


def get_one_all_dict(prediction, labels):
    one_idx = np.where(labels == 1)[0]
    unique, counts = np.unique(prediction, return_counts=True)
    all_dict = {unique[i]: counts[i] for i in range(len(unique))}
    all_counts = [all_dict.get(i, 0) for i in range(len(unique))]

    one_prediction = prediction[one_idx, ]
    unique, counts = np.unique(one_prediction, return_counts=True)
    one_dict = {unique[i]: counts[i] for i in range(len(unique))}
    one_counts = [one_dict.get(i, 0) for i in range(len(all_counts))]
    return one_dict, all_dict



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Max sampling while saving prediction probabilities.
'''

from modAL.utils.selection import multi_argmax
from modAL.utils.selection import shuffled_argmax

from asreview.query_strategies.base import ProbaQueryStrategy


class MaxQuery(ProbaQueryStrategy):
    "Maximum sampling query strategy."
    name = "max"

    def __init__(self, random_tie_break=False):
        """Initialize the max query strategy.

        Arguments:
        ----------
        random_tie_break: bool
            If true randomly decide which ones to include by tie-break.
        """
        super(MaxQuery, self).__init__()
        self.random_tie_break = random_tie_break

    def _query(self, X, pool_idx, n_instances=1, proba=None):
        proba = proba[pool_idx]
        if not self.random_tie_break:
            query_idx = multi_argmax(proba[:, 1], n_instances=n_instances)
        else:
            query_idx = shuffled_argmax(proba[:, 1], n_instances=n_instances)

        return pool_idx[query_idx], X[pool_idx[query_idx]]



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from PyInquirer import prompt, Separator

from asreview.config import NOT_AVAILABLE
from asreview.review import BaseReview
from asreview.types import convert_list_type
from asreview.logging.utils import open_logger


def update_stats(stats, label):
    if label == 1:
        stats["n_included"] += 1
        stats["last_inclusion"] = 0
    else:
        stats["n_excluded"] += 1
        stats["last_inclusion"] += 1
    stats["n_reviewed"] += 1
    stats["n_pool"] -= 1


class ReviewOracle(BaseReview):
    """ Review class for Oracle mode on the command line. """

    def __init__(self, X, as_data, *args, use_cli_colors=True,
                 **kwargs):
        self.as_data = as_data
        super(ReviewOracle, self).__init__(
            X,
            y=np.tile([NOT_AVAILABLE], X.shape[0]),
            *args,
            **kwargs)

        self.use_cli_colors = use_cli_colors

    def _papers_from_finder(self, logger):
        "Find papers using a fuzzy finder in the available records."
        question = [
            {
                'type': 'input',
                'name': 'keywords',
                'message': 'Find papers using keywords/authors/title:',
            }
        ]
        try:
            keywords = prompt(question)['keywords']
        except KeyError:
            return

        paper_idx = self.as_data.fuzzy_find(keywords, exclude=self.train_idx)

        # Get the (possibly) relevant papers.
        choices = []
        for idx in paper_idx:
            choices.append(self.as_data.preview_record(idx))
        choices.extend([Separator(), "return"])

        # Stay in the same menu until no more options are left
        while len(choices) > 2:
            question = [
                {
                    'type': 'list',
                    'name': 'paper',
                    'message': 'Choose a paper to review:',
                    'choices': choices
                }
            ]
            new_choice = prompt(question).get('paper', "return")
            if new_choice == "return":
                return
            choice_idx = choices.index(new_choice)
            idx = paper_idx[choice_idx]

            # Get the label for the selected paper.
            label = self._get_labels_paper(idx, ask_stop=False)
            self.classify([idx], [label], logger, method="initial")

            # Remove the selected choice from the list.
            del choices[choice_idx]
            del paper_idx[choice_idx]
        return

    def _papers_from_id(self, logger):
        "Get papers by a list of IDs."
        question = [
            {
                'type': 'input',
                'name': 'included',
                'message': 'Which papers do you want to include?\n'
                'Separate paper indices by spaces:',
            },
            {
                'type': 'input',
                'name': 'excluded',
                'message': 'Which papers do you want to exclude?\n'
                'Separate paper indices by spaces:',
            }
        ]
        answer = prompt(question)
        try:
            included = answer["included"]
            excluded = answer["excluded"]
        except KeyError:
            return

        new_included = convert_list_type(included.split(), int)
        new_excluded = convert_list_type(excluded.split(), int)
        self.classify(new_included, np.ones(len(new_included)),
                      logger, method="initial")
        self.classify(new_excluded, np.zeros(len(new_excluded)),
                      logger, method="initial")

    def main_menu(self, logger, *args, **kwargs):
        "Get initial papers for modelling."
        while True:
            n_included = np.sum(self.y[self.train_idx] == 1)
            n_excluded = np.sum(self.y[self.train_idx] == 0)
            question = [{
                'type': 'list',
                'name': 'action',
                'message': 'What do you want to do next?',
                'choices': [
                    "Find papers by keywords",
                    "Find papers by ID",
                    Separator(),
                    f"Continue review ({n_included} included, "
                    f"{n_excluded} excluded)",
                    "Export",
                    Separator(),
                    "Stop"
                ]
            }]
            action = prompt(question).get("action", "Stop")

            if action.endswith("by keywords"):
                self._papers_from_finder(logger)
            elif action.endswith("by ID"):
                self._papers_from_id(logger)
            elif action.startswith("Export"):
                self._export()
            elif action.startswith("Stop"):
                question = [{
                    'type': 'confirm',
                    'message': "Are you sure you want to stop?",
                    'name': 'stop',
                    'default': 'false',
                }]
                stop = prompt(question).get('stop', True)
                if stop:
                    raise KeyboardInterrupt
            elif action.startswith("Continue review"):
                try:
                    self._do_review(logger, *args, **kwargs)
                except KeyboardInterrupt:
                    pass

    def review(self, *args, instant_save=True, **kwargs):
        with open_logger(self.log_file) as logger:
            self.main_menu(logger, *args, instant_save=instant_save, **kwargs)

    def _format_paper(self,
                      title=None,
                      abstract=None,
                      keywords=None,
                      authors=None):

        if self.use_cli_colors:
            title = "\033[95m" + title + "\033[0m"

        return f"\n{title}\n{authors}\n\n{abstract}\n"

    def _get_labels_paper(self, index, stat_str=None, ask_stop=False):
        """Ask the user for a label for a particular paper.

        Arguments
        ---------
        index: int
            Paper ID in the dataset.
        stat_str: str
            Display this (statistic) string under the paper.
        ask_stop: bool
            Ask for confirmation when stopping.
        """
        # CLI paper format
        self.as_data.print_record(index)
        if stat_str is not None:
            print(stat_str + "\n")

        def _interact():
            question = [
                {
                    'type': 'list',
                    'name': 'action',
                    'message': 'Include or Exclude?',
                    'default': 'Exclude',
                    'choices': [
                        'Exclude', 'Include', Separator(), 'Back to main menu'
                    ],
                    'filter': lambda val: val.lower()
                }
            ]
            action = prompt(question).get("action", 'Back to main menu')
            return action

        action = _interact()

        if action == "include":
            label = 1
        elif action == "exclude":
            label = 0
        else:
            label = None

        return label

    def train(self, *args, **kwargs):
        super(ReviewOracle, self).train(*args, **kwargs)

    def _export(self):
        """Export the results to a csv file.

        Order of records is: [included, not reviewed (by proba), excluded]
        """
        question = [
            {
                'type': 'input',
                'name': 'file_name',
                'validate': lambda val: val.endswith((".csv", ".ris")),
                'message': 'Type file name for export ending with .csv or .ris'
            }
        ]
        try:
            file_name = prompt(question)["file_name"]
        except KeyError:
            return
        pred_proba = self.query_kwargs.get('pred_proba', None)
        pool_idx = np.delete(np.arange(len(self.y)), self.train_idx)
        if pred_proba is not None:
            proba_order = np.argsort(-pred_proba[pool_idx, 1])
        else:
            proba_order = np.arange(len(pool_idx))
        train_zero = self.train_idx[np.where(self.y[self.train_idx] == 0)[0]]
        train_one = self.train_idx[np.where(self.y[self.train_idx] == 1)[0]]
        df_order = np.concatenate(
            (train_one, pool_idx[proba_order], train_zero), axis=None)
        labels = np.full(len(self.y), np.nan, dtype=object)
        labels[self.train_idx] = self.y[self.train_idx]
        self.as_data.to_file(fp=file_name, labels=labels,
                             df_order=df_order)

    def get_stats(self, stats):
        n_included = stats["n_included"]
        n_papers = stats["n_papers"]
        n_reviewed = stats["n_reviewed"]
        perc_read = 100*(stats["n_reviewed"]/stats["n_papers"])
        if(n_reviewed == 0):
            perc_included = np.nan
        else:
            perc_included = 100*n_included/n_reviewed
        last_inclusion = stats["last_inclusion"]
        stat_str = (f"| {perc_read:.2f}% read | {last_inclusion} since last "
                    f"inclusion | {perc_included:.2f}% included |"
                    f" total papers: {n_reviewed}/{n_papers} |")
        return stat_str

    def _get_labels(self, ind):
        "Get a sequence of labels."
        y = np.zeros((len(ind), ), dtype=np.int)
        stats = self.statistics()

        for j, index in enumerate(ind):
            label = self._get_labels_paper(index,
                                           stat_str=self.get_stats(stats))
            if label is None:
                raise KeyboardInterrupt
            update_stats(stats, label)
            y[j] = label

        return y



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.init_sampling import sample_prior_knowledge
from asreview.review import BaseReview
from asreview.review.base import _merge_prior_knowledge


class ReviewSimulate(BaseReview):
    """ Automated Systematic Review in simulation mode. """

    def __init__(self,
                 X,
                 y,
                 *args,
                 n_prior_included=None,
                 n_prior_excluded=None,
                 **kwargs):
        self.n_prior_included = n_prior_included
        self.n_prior_excluded = n_prior_excluded
        super(ReviewSimulate, self).__init__(
            X, y, *args, **kwargs)

    def _prior_knowledge(self, logger):
        """ Get the prior knowledge, either from specific paper IDs,
            and if they're not given from the number of in/exclusions. """
        if self.prior_included is not None or self.prior_excluded is not None:
            prior_indices, prior_labels = _merge_prior_knowledge(
                self.prior_included,
                self.prior_excluded
            )
        else:
            # Create the prior knowledge
            init_ind = sample_prior_knowledge(
                self.y,
                n_prior_included=self.n_prior_included,
                n_prior_excluded=self.n_prior_excluded,
                random_state=None  # TODO
            )
            prior_indices, prior_labels = init_ind, self.y[init_ind, ]
        self.classify(prior_indices, prior_labels, logger, method="initial")

    def _get_labels(self, ind):
        """ Get the labels directly from memory.

        Arguments
        ---------
        ind: list, np.array
            A list with indices

        Returns
        -------
        list, np.array
            The corresponding true labels for each indice.
        """

        return self.y[ind, ]



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import ceil

import numpy as np

from asreview.balance_strategies.base import BaseBalance


class UndersampleBalance(BaseBalance):
    """Balancing class that undersamples the data with a given ratio.
    """
    name = "undersampling"

    def __init__(self, ratio=1.0):
        """Initialize the undersampling balance strategy.

        Arguments
        ---------
        ratio: double
            Undersampling ratio of the zero's. If for example we set a ratio of
            0.25, we would sample only a quarter of the zeros and all the ones.
        """
        super(UndersampleBalance, self).__init__()
        self.ratio = ratio

    def sample(self, X, y, train_idx, shared):
        one_ind = train_idx[np.where(y[train_idx] == 1)]
        zero_ind = train_idx[np.where(y[train_idx] == 0)]

        n_one = len(one_ind)
        n_zero = len(zero_ind)

        # If we don't have an excess of 0's, give back all training_samples.
        if n_one/n_zero >= self.ratio:
            shuf_ind = np.append(one_ind, zero_ind)
        else:
            n_zero_epoch = ceil(n_one/self.ratio)
            zero_under = np.random.choice(np.arange(n_zero), n_zero_epoch,
                                          replace=False)
            shuf_ind = np.append(one_ind, zero_ind[zero_under])

        np.random.shuffle(shuf_ind)
        return X[shuf_ind], y[shuf_ind]

    def full_hyper_space(self):
        from hyperopt import hp
        parameter_space = {
            "bal_ratio": hp.lognormal('bal_ratio', 0, 1),
        }
        return parameter_space, {}



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_query_class(method):
    """Get class of query strategy from its name.


    Arguments
    ---------
    method: str
        Name of the query strategy, e.g. 'max', 'uncertainty', 'random.
        A special mixed query strategy is als possible. The mix is denoted
        by an underscore: 'max_random' or 'max_uncertainty'.

    Returns
    -------
    BaseQueryModel:
        Class corresponding to the method name.
    """
    from asreview.query_strategies.cluster import ClusterQuery
    from asreview.query_strategies.max import MaxQuery
    from asreview.query_strategies.uncertainty import UncertaintyQuery
    from asreview.query_strategies.random import RandomQuery
    from asreview.query_strategies.mixed import MixedQuery
    query_models = {
        "cluster": ClusterQuery,
        "max": MaxQuery,
        "uncertainty": UncertaintyQuery,
        "random": RandomQuery,
    }

    # Try to split the query strategy if the string wasn't found.
    try:
        return query_models[method]
    except KeyError:
        mix = method.split("_")
        if len(mix) == 2:
            return MixedQuery
        raise ValueError(f"Error: query method '{method}' is not implemented.")


def get_query_model(method, *args, **kwargs):
    """Get an instance of the query strategy.

    Arguments
    ---------
    method: str
        Name of the query strategy.
    *args:
        Arguments for the model.
    **kwargs:
        Keyword arguments for the model.

    Returns
    -------
    BaseQueryModel:
        Initialized instance of query strategy.
    """
    from asreview.query_strategies.mixed import MixedQuery
    query_class = get_query_class(method)
    if query_class == MixedQuery:
        mix = method.split("_")
        for i in range(2):
            kwargs.pop("strategy_" + str(i+1), None)
        return query_class(mix[0], mix[1], *args, **kwargs)
    return query_class(*args, **kwargs)



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from math import floor

import numpy as np
from scipy.sparse import issparse
from scipy.sparse import vstack

from asreview.query_strategies.base import BaseQueryStrategy
from asreview.query_strategies.utils import get_query_model


class MixedQuery(BaseQueryStrategy):
    """Class for mixed query strategy.

    The idea is to use two different query strategies at the same time with a
    ratio of one to the other.
    """
    name = "mixed"

    def __init__(self, strategy_1="max", strategy_2="random", mix_ratio=0.95,
                 **kwargs):
        """Initialize the Mixed query strategy

        Arguments
        ---------
        strategy_1: str
            Name of the first query strategy.
        strategy_2: str
            Name of the second query strategy.
        mix_ratio: float
            Portion of queries done by the first strategy. So a mix_ratio of
            0.95 means that 95% of the time query strategy 1 is used and 5% of
            the time query strategy 2.
        **kwargs: dict
            Keyword arguments for the two strategy. To specify which of the
            strategies the argument is for, prepend with the name of the query
            strategy and an underscore, e.g. 'max_' for maximal sampling.
        """
        super(MixedQuery, self).__init__()
        kwargs_1 = {}
        kwargs_2 = {}
        for key, value in kwargs.items():
            if key.startswith(strategy_1):
                new_key = key[len(strategy_1)+1:]
                kwargs_1[new_key] = value
            elif key.starts_with(strategy_2):
                new_key = key[len(strategy_2)+1:]
                kwargs_2[new_key] = value
            else:
                logging.warn(f"Key {key} is being ignored for the mixed "
                             "({strategy_1}, {strategy_2}) query strategy.")

        self.query_model1 = get_query_model(strategy_1)
        self.query_model2 = get_query_model(strategy_2)
        self.mix_ratio = mix_ratio

    def query(self, X, classifier, pool_idx=None, n_instances=1, shared={}):
        n_samples = X.shape[0]
        if pool_idx is None:
            pool_idx = np.arange(n_samples)

        # Split the number of instances for the query strategies.
        n_instances_1 = floor(n_instances*self.mix_ratio)
        leftovers = n_instances*self.mix_ratio-n_instances_1
        if np.random.random_sample() < leftovers:
            n_instances_1 += 1
        n_instances_2 = n_instances-n_instances_1

        # Perform the query with strategy 1.
        query_idx_1, X_1 = self.query_model1.query(
            X, classifier, pool_idx=pool_idx, n_instances=n_instances_1,
            shared=shared)

        # Remove the query indices from the pool.
        train_idx = np.delete(np.arange(n_samples), pool_idx, axis=0)
        train_idx = np.append(train_idx, query_idx_1)
        new_pool_idx = np.delete(np.arange(n_samples), train_idx, axis=0)

        # Perform the query with strategy 2.
        query_idx_2, X_2 = self.query_model2.query(
            X, classifier, pool_idx=new_pool_idx, n_instances=n_instances_2,
            shared=shared)

        query_idx = np.append(query_idx_1, query_idx_2)

        if n_instances_1 == 0:
            X = X_2
        elif n_instances_2 == 0:
            X = X_1
        else:
            if issparse(X_1) and issparse(X_2):
                X = vstack([X_1, X_2]).tocsr()
            else:
                X = np.concatenate((X_1, X_2), axis=0)

        return query_idx, X

    def full_hyper_space(self):
        from hyperopt import hp

        space_1, choices_1 = self.query_model1.hyper_space()
        space_2, choices_2 = self.query_model2.hyper_space()
        parameter_space = {}
        hyper_choices = {}
        for key, value in space_1.items():
            new_key = "qry_" + self.strategy_1 + key[4:]
            parameter_space[new_key] = value
            hyper_choices[new_key] = choices_1[key]

        for key, value in space_2.items():
            new_key = "qry_" + self.strategy_2 + key[4:]
            parameter_space[new_key] = value
            hyper_choices[new_key] = choices_2[key]

        parameter_space["qry_mix_ratio"] = hp.uniform(
            "qry_mix_ratio", 0, 1)

        return parameter_space, hyper_choices



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Functions that resample the training data.
'''
import numpy as np


def validation_data(X, y, fit_kwargs, ratio=1):
    """ Set validation data from the samples remaining in the pool. """
    one_ind = np.where(y == 1)[0]
    zero_ind = np.where(y == 0)[0]
    n_one = len(one_ind)
    n_zero = len(zero_ind)

    n_zero_add = min(n_zero, int(ratio*n_one))

    zero_add = zero_ind[np.random.choice(n_zero, n_zero_add, replace=False)]
    all_ind = np.append(one_ind, zero_add)
    fit_kwargs['validation_data'] = (X[all_ind], y[all_ind])



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from sklearn.linear_model import LogisticRegression

from asreview.models.base import BaseTrainModel


class LogisticModel(BaseTrainModel):
    "Logistic Regression SKLearn model."
    name = "logistic"

    def __init__(self, C=1.0, class_weight=None, n_jobs=1):
        """Initialize the SKLearn Naive Bayes model.

        Arguments:
        ----------
        alpha: float
            Parameter to set the regularization strength of the model.
        """
        super(LogisticModel, self).__init__()
        if class_weight is not None:
            class_weight = {
                0: 1,
                1: class_weight,
            }
        self.C = C
        self.class_weight = class_weight
        self.n_jobs = n_jobs
        self._model = LogisticRegression(C=C, class_weight=class_weight,
                                         n_jobs=n_jobs)
        logging.debug(self._model)

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {}
        hyper_space = {
            "mdl_C": hp.lognormal("mdl_C", 0, 1),
            "mdl_class_weight": hp.lognormal("mdl_class_weight", 0, 1),
        }
        return hyper_space, hyper_choices



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from pathlib import Path


def _unsafe_dict_update(default_dict, override_dict):
    """
    Using defaults and an overriding dictionary, create a new dictionary.
    This new dictionary has the same values as the default dictionary and
    the same types. Thus, if there are values that are in the overriding
    dictionary, but not in the original, they will be ignored.

    Arguments
    ---------
    default_dict: dict
        Starting dictionary with defaults.
    override_dict: dict
        Dictionary with custom values (such as model parameters).

    Returns
    -------
    dict
        Merged dictionary.
    """
    new_dict = default_dict
    for key in override_dict:
        if key not in default_dict:
            print(f"Warning: key {key} is being ignored.")

    for key in new_dict:
        if key in override_dict:
            str_val = override_dict[key]
            if type(new_dict[key]) == bool:
                new_dict[key] = str_val in ["True", "true", "T", "t", True]
            else:
                try:
                    new_dict[key] = type(new_dict[key])(str_val)
                except TypeError:
                    raise(TypeError(f"Error at {key}"))
    return new_dict


def _safe_dict_update(default_dict, override_dict):
    """
    Using defaults and an overriding dictionary, create a new dictionary.
    This new dictionary has the same values as the default dictionary.
    Thus, if there are values that are in the overriding
    dictionary, but not in the original, they will be ignored.
    In contrast to the unsafe version, the type should be supplied in the
    default dictionary: key: (value, type).

    Arguments
    ---------
    default_dict: dict
        Starting dictionary with defaults.
    override_dict: dict
        Dictionary with custom values (such as model parameters).

    Returns
    -------
    dict
        Merged dictionary.
    """
    new_dict = {}
    for key in default_dict:
        new_dict[key] = default_dict[key][0]

    for key in override_dict:
        if key not in default_dict:
            print(f"Warning: key {key} is being ignored.")

    for key in new_dict:
        if key in override_dict:
            str_val = override_dict[key]
            type_val = default_dict[key][1]
            if type_val == bool:
                new_dict[key] = str_val in ["True", "true", "T", "t"]
            else:
                try:
                    new_dict[key] = type_val(str_val)
                except TypeError:
                    raise(TypeError(f"Error at {key}"))
    return new_dict


def get_data_home(data_home=None):
    """Return the path of the ASR data dir.

    This folder is used by some large dataset loaders to avoid downloading the
    data several times.
    By default the data dir is set to a folder named 'asr_data' in the
    user home folder.
    Alternatively, it can be set by the 'ASR_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home : str | None
        The path to scikit-learn data dir.

    """
    if data_home is None:
        data_home = os.environ.get('ASR_DATA',
                                   Path('~', 'asr_data'))
    data_home = Path(data_home).expanduser()

    if not data_home.exists():
        data_home.mkdir(parents=True, exist_ok=True)

    return data_home


def _set_class_weight(weight1):
    """ Used in RNN's to have quicker learning. """
    weight0 = 1.0
    cw_class = {
        0: weight0,
        1: weight1,
    }
    logging.debug(f"Using class weights: 0 <- {weight0}, 1 <- {weight1}")
    return cw_class



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Statistical functions for finding several different measures."
import numpy as np


def _find_inclusions(logger, labels, remove_initial=True):
    "Find the number of inclusions found at each step."
    inclusions = []
    n_initial_inc = 0
    cur_inclusions = 0
    n_initial = 0
    n_queries = logger.n_queries()
    for query_i in range(n_queries):
        label_methods = logger.get("label_methods", query_i)
        label_idx = logger.get("label_idx", query_i)
        for i in range(len(label_idx)):
            if label_methods[i] == "initial" and remove_initial:
                n_initial_inc += labels[label_idx[i]]
                n_initial += 1
            else:
                cur_inclusions += labels[label_idx[i]]
                inclusions.append(cur_inclusions)

    inclusions_after_init = sum(labels)
    if remove_initial:
        inclusions_after_init -= n_initial_inc
    return inclusions, inclusions_after_init, n_initial


def _get_labeled_order(logger):
    "Get the order in which papers were labeled."
    label_order = []
    n_initial = 0
    n_queries = logger.n_queries()
    for query_i in range(n_queries):
        try:
            label_methods = logger.get("label_methods", query_i)
        except KeyError:
            continue
        label_idx = logger.get("label_idx", query_i)
        for i in range(len(label_idx)):
            if label_methods[i] == "initial":
                n_initial += 1
        label_order.extend(label_idx)
    return label_order, n_initial


def _get_last_proba_order(logger):
    "Get the ranking of papers in the last query."
    n_queries = logger.n_queries()
    pool_idx = None
    for query_i in reversed(range(n_queries)):
        try:
            pool_idx = logger.get("pool_idx", query_i)
        except KeyError:
            continue
        if pool_idx is not None:
            proba = logger.get("proba", query_i)
            break

    if pool_idx is None:
        return []
    return pool_idx[np.argsort(-proba[pool_idx])]


def _get_proba_order(logger, query_i):
    "Get the ranking of papers in query_i."
    try:
        pool_idx = logger.get("pool_idx", query_i)
    except KeyError:
        pool_idx = None

    if pool_idx is None:
        return None

    proba = logger.get("proba", query_i)[pool_idx]
    return pool_idx[np.argsort(proba)]


def _n_false_neg(logger, query_i, labels):
    "Find the number of false negatives after reading x papers."
    proba_order = _get_proba_order(logger, query_i)
    if proba_order is None:
        return None
    res = np.zeros(len(proba_order))

    n_one = 0
    for i in range(len(res)):
        if labels[proba_order[i]] == 1:
            n_one += 1
        res[i] = n_one
    return np.array(list(reversed(res)))


def _get_limits(loggers, query_i, labels, proba_allow_miss=[]):
    "Get the number of papers to be read, with a criterium."
    num_left = None

    for logger in loggers.values():
        new_num_left = _n_false_neg(logger, query_i, labels)
        if new_num_left is None:
            return None

        if num_left is None:
            num_left = new_num_left
        else:
            num_left += new_num_left
    num_left /= len(loggers)
    limits = [len(num_left)]*len(proba_allow_miss)
    allow_miss = {i: proba for i, proba in enumerate(proba_allow_miss)}
    for i in range(len(num_left)):
        for i_prob, prob in list(allow_miss.items()):
            if num_left[i] < prob:
                limits[i_prob] = i
                del allow_miss[i_prob]
        if len(allow_miss) == 0:
            break
    return limits


def _random_ttd(loggers, query_i, labels):
    all_ttd = []
    for logger in loggers.values():
        try:
            pool_idx = logger.get("pool_idx", query_i)
        except KeyError:
            continue

        pool_labels = labels[pool_idx]
        n_included = np.sum(pool_labels)

        if n_included == 0:
            continue
        ttd = 0
        p_only_zero = 1
        for i in range(len(pool_labels) - n_included):
            p_first = n_included/(len(pool_labels)-i)
            ttd += p_only_zero*p_first*i
            p_only_zero *= (1-p_first)
        all_ttd.append(ttd)

    if len(all_ttd) == 0:
        ttd_avg = 0
    else:
        ttd_avg = np.average(all_ttd)
    return ttd_avg


def _max_ttd(loggers, query_i, labels):
    all_ttd = []
    for logger in loggers.values():
        proba_order = _get_proba_order(logger, query_i)
        if proba_order is None:
            all_ttd.append(0)
            continue
        if len(proba_order) == 0:
            continue

        x = np.where(labels[proba_order] == 1)[0]
        if len(x) == 0:
            ttd = 0
        else:
            ttd = (len(proba_order)-1) - x[-1]
        all_ttd.append(ttd)

    if len(all_ttd) == 0:
        ttd_avg = 0
    else:
        ttd_avg = np.average(all_ttd)
    return ttd_avg


def _cluster_order(all_dict, power=0):
    scores = []
    for clust_id in all_dict:
        for i in range(all_dict[clust_id]):
            new_score = (i+1) * pow(all_dict[clust_id], -power)
            scores.append((clust_id, new_score))
    scores = sorted(scores, key=lambda x: x[1])
    return [x[0] for x in scores]


def _get_clustering(all_prediction, pool_idx, labels):
    pool_prediction = all_prediction[pool_idx]
    one_idx = np.where(labels[pool_idx] == 1)[0]
    unique, counts = np.unique(pool_prediction, return_counts=True)
    all_dict = {unique[i]: counts[i] for i in range(len(unique))}
    all_counts = [all_dict.get(i, 0) for i in range(np.max(unique)+1)]

    prediction = pool_prediction[one_idx, ]
    unique, counts = np.unique(prediction, return_counts=True)
    one_dict = {unique[i]: counts[i] for i in range(len(unique))}
    one_counts = [one_dict.get(i, 0) for i in range(len(all_counts))]
    return all_dict, all_counts, one_dict, one_counts


def _cluster_ttd(loggers, query_i, labels, all_prediction):
    all_ttd = []
    for logger in loggers.values():
        try:
            pool_idx = logger.get("pool_idx", query_i)
        except KeyError:
            all_ttd.append(0)
            continue

        all_dict, all_counts, _one_dict, one_counts = _get_clustering(
            all_prediction, pool_idx, labels)
        cluster_order = _cluster_order(all_dict)

        p_only_zero = 1
        ttd = 0
        if np.sum(one_counts) == 0:
            continue
        for i, i_clust in enumerate(cluster_order):
            try:
                p_first = one_counts[i_clust]/all_counts[i_clust]
            except IndexError:
                print(
                    i_clust, list(all_dict), len(all_counts), len(one_counts))
            ttd += p_only_zero*p_first*i
            p_only_zero *= 1-p_first
            all_counts[i_clust] -= 1
            if p_only_zero < 1e-6:
                break
        all_ttd.append(ttd)

    if len(all_ttd) == 0:
        ttd_avg = 0
    else:
        ttd_avg = np.average(all_ttd)
    return ttd_avg



from pytest import mark
import numpy as np

from asreview.balance_strategies.utils import get_balance_model


def generate_data(n_feature=20, n_sample=10):
    X = np.random.rand(n_sample, n_feature)
    n_sample_zero = np.int(n_sample/2)
    n_sample_one = n_sample - n_sample_zero
    y = np.append(np.zeros(n_sample_zero), np.ones(n_sample_one))
    np.random.shuffle(y)

    return X, y


def check_partition(X, y, X_partition, y_partition, train_idx):
    partition_idx = []
    for row in X_partition:
        partition_idx.append(
            np.where(np.all(X == row, axis=1))[0][0]
        )

    assert np.count_nonzero(y_partition == 0) > 0
    assert np.count_nonzero(y_partition == 1) > 0
    assert len(partition_idx) == X_partition.shape[0]
    assert set(partition_idx) <= set(train_idx.tolist())
    assert np.all(X[partition_idx] == X_partition)
    assert np.all(y[partition_idx] == y_partition)


@mark.parametrize(
    "balance_strategy",
    [
        "undersample",
        "simple",
        "double",
        "triple",
    ])
def test_balance(balance_strategy, n_partition=100, n_feature=200,
                 n_sample=100):
    model = get_balance_model(balance_strategy)
    X, y = generate_data(n_feature=n_feature, n_sample=n_sample)
    for _ in range(n_partition):
        n_train = np.random.randint(10, n_sample)
        while True:
            train_idx = np.random.choice(
                np.arange(len(y)), n_train, replace=False)
            num_zero = np.count_nonzero(y[train_idx] == 0)
            num_one = np.count_nonzero(y[train_idx] == 1)
            if num_zero > 0 and num_one > 0:
                break
        shared = {"query_src": {}, "current_queries": {}}
        X_train, y_train = model.sample(X, y, train_idx, shared)
        check_partition(X, y, X_train, y_train, train_idx)



import os

from pytest import mark

from asreview.feature_extraction.utils import get_feature_model
from asreview.readers import ASReviewData


@mark.parametrize(
    "feature_extraction",
    [
        "doc2vec",
        "embedding-idf",
        #  "sbert",
        "tfidf",
    ])
@mark.parametrize(
    "split_ta",
    [
        0,
        1,
    ])
def test_features(feature_extraction, split_ta):
    embedding_fp = os.path.join("test", "demo_data", "generic.vec")
    data_fp = os.path.join("test", "demo_data", "generic.csv")

    as_data = ASReviewData.from_file(data_fp)
    texts = as_data.texts
    if feature_extraction.startswith("embedding-"):
        model = get_feature_model(feature_extraction, split_ta=split_ta,
                                  embedding_fp=embedding_fp)
    else:
        model = get_feature_model(feature_extraction, split_ta=split_ta)
    X = model.fit_transform(texts, titles=as_data.title,
                            abstracts=as_data.abstract)

    assert X.shape[0] == len(as_data.title)
    assert X.shape[1] > 0



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.logging.utils import open_logger
from asreview.feature_extraction.embedding_lstm import load_embedding
from asreview.feature_extraction.embedding_lstm import sample_embedding
from asreview.feature_extraction.embedding_lstm import text_to_features
from asreview.readers import ASReviewData
from asreview.readers import read_csv
from asreview.readers import read_data
from asreview.readers import read_excel
from asreview.readers import read_pubmed_xml
from asreview.readers import read_ris
from asreview.review import get_reviewer
from asreview.review import MinimalReview
from asreview.review import review
from asreview.review import review_oracle
from asreview.review import review_simulate
from asreview.review import ReviewOracle
from asreview.review import ReviewSimulate

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from pathlib import Path
import warnings
import xml.etree.ElementTree as ET


import numpy as np
import pandas as pd
from RISparser import readris
from RISparser import TAG_KEY_MAPPING

from asreview.config import NOT_AVAILABLE
from asreview.exceptions import BadFileFormatError

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    from fuzzywuzzy import fuzz

RIS_KEY_LABEL_INCLUDED = "LI"
NAME_LABEL_INCLUDED = "label_included"
LABEL_INCLUDED_VALUES = [
    "label_included",
    "included_label",
    "included_final",
    "included",
    "included_flag"
]


def format_to_str(obj):
    res = ""
    if isinstance(obj, list):
        for sub in obj:
            res += str(sub) + " "
    else:
        res = obj
    return res


def _tag_key_mapping(reverse=False):
    # Add label_included into the specification and create reverse mapping.
    TAG_KEY_MAPPING[RIS_KEY_LABEL_INCLUDED] = NAME_LABEL_INCLUDED
    KEY_TAG_MAPPING = {TAG_KEY_MAPPING[key]: key for key in TAG_KEY_MAPPING}
    for label in LABEL_INCLUDED_VALUES:
        KEY_TAG_MAPPING[label] = "LI"
    if reverse:
        return KEY_TAG_MAPPING
    else:
        return TAG_KEY_MAPPING


def get_fuzzy_ranking(keywords, str_list):
    rank_list = np.zeros(len(str_list), dtype=np.float)
    for i, my_str in enumerate(str_list):
        rank_list[i] = fuzz.token_set_ratio(keywords, my_str)
    return rank_list


def merge_arrays(array_a, array_b, n_paper_a, n_paper_b, fill="", type_=object):
    if array_a is None and array_b is not None:
        array_a = np.full(n_paper_a, fill).astype(type_)
    if array_a is not None and array_b is None:
        array_b = np.full(n_paper_b, fill).astype(type_)
    if array_a is not None:
        array_a = np.append(array_a, array_b)
    return array_a


class ASReviewData(object):
    """Data object to store csv/ris file.

    Extracts relevant properties of papers. """
    def __init__(self, raw_df, labels=None, title=None, abstract=None,
                 keywords=None, article_id=None, authors=None,
                 label_col=LABEL_INCLUDED_VALUES[0], final_labels=None):
        self.raw_df = raw_df
        self.labels = labels
        self.title = title
        self.abstract = abstract
        self.label_col = label_col
        self.keywords = keywords
        self.article_id = article_id
        self.authors = authors
        self.final_labels = final_labels

        if authors is None:
            logging.warning("Could not locate authors in data.")
        if title is None:
            BadFileFormatError("Could not locate titles in data.")
        if abstract is None:
            BadFileFormatError("Could not locate abstracts in data.")

        self.n_paper_train = len(self.raw_df.index)
        self.n_paper = self.n_paper_train
        if article_id is None:
            self.article_id = np.arange(len(raw_df.index))

    def append(self, as_data):
        """Append another ASReviewData object.

        It puts the training data at the end.

        Arguments
        ---------
        as_data: ASReviewData
            Dataset to append.
        """
        if as_data.labels is None:
            BadFileFormatError("Additional datasets should have labels.")
        if self.labels is None:
            self.labels = np.full(self.n_paper, NOT_AVAILABLE)
        self.labels = np.append(self.labels, as_data.labels)

        self.title = np.append(self.title, as_data.title)
        self.abstract = np.append(self.abstract, as_data.abstract)
        self.article_id = np.append(self.article_id, as_data.article_id + self.n_paper)
        self.keywords = merge_arrays(self.keywords, as_data.keywords, self.n_paper,
                                     as_data.n_paper, "", object)
        self.authors = merge_arrays(self.authors, as_data.authors, self.n_paper,
                                    as_data.n_paper, "", object)
        self.final_labels = merge_arrays(self.final_labels, as_data.final_labels,
                                         self.n_paper, as_data.n_paper, NOT_AVAILABLE,
                                         np.int)
        self.n_paper += as_data.n_paper

    @classmethod
    def from_data_frame(cls, raw_df, abstract_only=False):
        """ Get a review data object from a pandas dataframe. """
        # extract the label column
        column_labels = [label for label in list(raw_df)
                         if label in LABEL_INCLUDED_VALUES]

        if len(column_labels) > 1:
            print('\x1b[0;30;41m Warning multiple valid label inclusion '
                  'columns detected. \x1b[0m')
            print(f'Possible values: {column_labels}.')
            print(f'Choosing the one with the highest priority: '
                  f'{column_labels[0]}')
        data_kwargs = {"raw_df": raw_df}

        if len(column_labels) > 0:
            data_kwargs['labels'] = np.array(raw_df[column_labels[0]].fillna(
                NOT_AVAILABLE).values, dtype=np.int)
            data_kwargs['label_col'] = column_labels[0]
        else:
            data_kwargs['label_col'] = LABEL_INCLUDED_VALUES[0]

        if 'inclusion_code' in raw_df.columns and abstract_only:
            inclusion_codes = raw_df['inclusion_code'].fillna(
                NOT_AVAILABLE).values
            inclusion_codes = np.array(inclusion_codes, dtype=np.int)
            data_kwargs['final_labels'] = data_kwargs['labels']
            data_kwargs['labels'] = inclusion_codes > 0

        def fill_column(dst_dict, keys):
            if not isinstance(keys, list):
                keys = [keys]
            dst_key = keys[0]
            df_columns = {str(col_name).lower(): col_name
                          for col_name in list(raw_df)}
            for key in keys:
                try:
                    dst_dict[dst_key] = raw_df[df_columns[key]].fillna('').values
                except KeyError:
                    pass

        for key in [['title', 'primary_title'],
                    ['authors', 'author names', 'first_authors'],
                    'abstract', 'keywords']:
            fill_column(data_kwargs, key)

        return cls(**data_kwargs)

    @classmethod
    def from_csv(cls, fp, *args, **kwargs):
        return cls.from_data_frame(pd.DataFrame(read_csv(fp)), *args, **kwargs)

    @classmethod
    def from_ris(cls, fp, *args, **kwargs):
        return cls.from_data_frame(pd.DataFrame(read_ris(fp)), *args, **kwargs)

    @classmethod
    def from_excel(cls, fp, *args, **kwargs):
        return cls.from_data_frame(
            pd.DataFrame(read_excel(fp)), *args, **kwargs)

    @classmethod
    def from_file(cls, fp, *args, extra_dataset=[], **kwargs):
        "Create instance from csv/ris/excel file."
        as_data = cls.from_data_frame(_df_from_file(fp), *args, **kwargs)

        if len(extra_dataset) == 0:
            return as_data

        for prior_fp in extra_dataset:
            prior_as_data = cls.from_data_frame(
                _df_from_file(prior_fp), *args, **kwargs)
            as_data.append(prior_as_data)
        return as_data

    def preview_record(self, i, w_title=80, w_authors=40):
        "Return a preview string for record i."
        title_str = ""
        author_str = ""
        if self.title is not None:
            if len(self.title[i]) > w_title:
                title_str = self.title[i][:w_title-2] + ".."
            else:
                title_str = self.title[i]
        if self.authors is not None:
            cur_authors = format_to_str(self.authors[i])
            if len(cur_authors) > w_authors:
                author_str = cur_authors[:w_authors-2] + ".."
            else:
                author_str = cur_authors
        format_str = "{0: <" + str(w_title) + "}   " + "{1: <" + str(w_authors)
        format_str += "}"
        prev_str = format_str.format(title_str, author_str)
        return prev_str

    def format_record(self, i, use_cli_colors=True):
        " Format one record for displaying in the CLI. "
        if self.title is not None and len(self.title[0]) > 0:
            title = self.title[i]
            if use_cli_colors:
                title = "\033[95m" + title + "\033[0m"
            title += "\n"
        else:
            title = ""

        if self.authors is not None and len(self.authors[i]) > 0:
            authors = format_to_str(self.authors[i]) + "\n"
        else:
            authors = ""

        if self.abstract is not None and len(self.abstract[i]) > 0:
            abstract = self.abstract[i]
            abstract = "\n" + abstract + "\n"
        else:
            abstract = ""

        return ("\n\n----------------------------------"
                f"\n{title}{authors}{abstract}"
                "----------------------------------\n\n")

    def print_record(self, *args, **kwargs):
        "Print a record to the CLI."
        print(self.format_record(*args, **kwargs))

    def fuzzy_find(self, keywords, threshold=50, max_return=10,
                   exclude=None):
        """Find a record using keywords.

        It looks for keywords in the title/authors/keywords
        (for as much is available). Using the fuzzywuzzy package it creates
        a ranking based on token set matching.

        Arguments
        ---------
        keywords: str
            A string of keywords together, can be a combination.
        threshold: float
            Don't return records below this threshold.
        max_return: int
            Maximum number of records to return.

        Returns
        -------
        list:
            Sorted list of indexes that match best the keywords.
        """
        match_str = np.full(self.title.shape, "x", dtype=object)

        if self.title is not None:
            for i, title in enumerate(self.title):
                match_str[i, ] = str(title) + " "
        if self.authors is not None:
            for i in range(len(self.authors)):
                match_str[i] += format_to_str(self.authors[i]) + " "
        if self.keywords is not None:
            if isinstance(self.keywords[0], list):
                new_keywords = np.array([" ".join(x) for x in self.keywords])
            else:
                new_keywords = self.keywords
            for i in range(len(new_keywords)):
                match_str[i] += str(new_keywords[i])

        new_ranking = get_fuzzy_ranking(keywords, match_str)
        sorted_idx = np.argsort(-new_ranking)
        best_idx = []
        for idx in sorted_idx:
            if idx in exclude:
                continue
            if len(best_idx) >= max_return:
                break
            if len(best_idx) > 0 and new_ranking[idx] < threshold:
                break
            best_idx.append(idx)
        return np.array(best_idx, dtype=np.int).tolist()

    @property
    def texts(self):
        return [self.title[i] + " " + self.abstract[i]
                for i in range(len(self.title))]

    def get_priors(self):
        "Get prior_included, prior_excluded from dataset."
        zero_idx = np.where(self.labels[self.n_paper_train:] == 0)[0]
        one_idx = np.where(self.labels[self.n_paper_train:] == 1)[0]
        return one_idx + self.n_paper_train, zero_idx + self.n_paper_train

    def to_file(self, fp, labels=None, df_order=None):
        """
        Export data object to file.
        Both RIS and CSV are supported file formats at the moment.

        Arguments
        ---------
        fp: str
            Filepath to export to.
        labels: list, np.array
            Labels to be inserted into the dataframe before export.
        df_order: list, np.array
            Optionally, dataframe rows can be reordered.
        """
        if Path(fp).suffix in [".csv", ".CSV"]:
            self.to_csv(fp, labels=labels, df_order=df_order)
        elif Path(fp).suffix in [".ris", ".RIS"]:
            self.to_ris(fp, labels=labels, df_order=df_order)
        else:
            raise ValueError(f"Unknown file extension: {Path(fp).suffix}.\n"
                             f"from file {fp}")

    def to_csv(self, csv_fp, labels=None, df_order=None):
        new_df = self.raw_df.copy()
        if labels is not None:
            new_df[self.label_col] = labels

        if df_order is not None:
            new_df = new_df.reindex(df_order)
        new_df.to_csv(csv_fp)

    def to_ris(self, ris_fp, labels=None, df_order=None):
        new_df = self.raw_df.copy()
        if labels is not None:
            new_df[self.label_col] = labels

        if df_order is not None:
            new_df = new_df.reindex(df_order)
        write_ris(new_df, ris_fp)


def write_ris(df, ris_fp):
    """Write dataframe to RIS file.

    Arguments
    ---------
    df: pandas.Dataframe
        Dataframe to export.
    ris_fp: str
        RIS file to export to.
    """
    column_names = list(df)
    column_key = []
    for col in column_names:
        try:
            rev_mapping = _tag_key_mapping(reverse=True)
            column_key.append(rev_mapping[col])
        except KeyError:
            column_key.append('UK')
            logging.info(f"Cannot find column {col} in specification.")

    n_row = df.shape[0]

    # According to RIS specifications, a record should begin with TY.
    # Thus, the column id is inserted before all the other.
    col_order = []
    for i, key in enumerate(column_key):
        if key == 'TY':
            col_order.insert(0, i)
        else:
            col_order.append(i)

    with open(ris_fp, "w") as fp:
        for i_row in range(n_row):
            for i_col in col_order:
                value = df.iloc[i_row, i_col]
                if isinstance(value, list):
                    for val in value:
                        fp.write(f"{column_key[i_col]}  - {val}\n")
                else:
                    fp.write(f"{column_key[i_col]}  - {value}\n")
            fp.write("ER  - \n\n")


def _df_from_file(fp):
    if Path(fp).suffix.lower() == ".csv":
        data = read_csv(fp)
    elif Path(fp).suffix.lower() in [".ris", ".txt"]:
        data = read_ris(fp)
    elif Path(fp).suffix.lower() == ".xlsx":
        data = read_excel(fp)
    elif Path(fp).suffix.lower() == ".xml":
        data = read_pubmed_xml(fp)
    else:
        raise ValueError(f"Unknown file extension: {Path(fp).suffix}.\n"
                         f"from file {fp}")

    # parse data in pandas dataframe
    df = pd.DataFrame(data)
    return df


def read_data(fp):
    """Load papers and their labels.

    Arguments
    ---------
    fp: str
        File path to the data.

    Returns
    -------
    np.ndarray, np.array
        The title and abstract merged into a single string for each paper.
        The labels for each paper. 1 is included, 0 is excluded. If this column
        is not available, this column is not returned.
    """
    df = _df_from_file(fp)
    # make texts
    texts = (df['title'].fillna('') + ' ' + df['abstract'].fillna(''))

    # extract the label column
    column_labels = [label for label in list(df)
                     if label in LABEL_INCLUDED_VALUES]

    if len(column_labels) > 1:
        print('\x1b[0;30;41m Warning multiple valid label inclusion '
              'columns detected. \x1b[0m')
        print(f'Possible values: {column_labels}.')
        print(f'Choosing the one with the highest priority: '
              f'{column_labels[0]}')
    elif len(column_labels) == 0:
        return df, texts.values, None
    labels = df[column_labels[0]].fillna(NOT_AVAILABLE)
    return df, texts.values, labels.values


def read_csv(fp):
    """CVS file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the CSV file.

    Returns
    -------
    list:
        List with entries.

    """
    try:
        df = pd.read_csv(fp)
    except UnicodeDecodeError:
        df = pd.read_csv(fp, encoding="ISO-8859-1")

    return df.to_dict('records')


def read_excel(fp):
    """Excel file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the Excel file (.xlsx).

    Returns
    -------
    list:
        List with entries.

    """
    try:
        dfs = pd.read_excel(fp, sheet_name=None)
    except UnicodeDecodeError:
        dfs = pd.read_excel(fp, sheet_name=None, encoding="ISO-8859-1")

    best_sheet = None
    sheet_obj_val = -1
    wanted_columns = [
        'title', 'primary_title', 'authors', 'author names', 'first_authors',
        'abstract', 'keywords']
    wanted_columns.extend(LABEL_INCLUDED_VALUES)
    for sheet_name in dfs:
        col_names = set([col.lower() for col in list(dfs[sheet_name])])
        obj_val = len(col_names & set(wanted_columns))
        if obj_val > sheet_obj_val:
            sheet_obj_val = obj_val
            best_sheet = sheet_name

    return dfs[best_sheet].to_dict('records')


def read_pubmed_xml(fp):
    """PubMed XML file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the XML file (.xml).

    Returns
    -------
    list:
        List with entries.
    """
    tree = ET.parse(fp)
    root = tree.getroot()

    records = []
    for child in root:
        parts = []
        elem = child.find('MedlineCitation/Article/ArticleTitle')
        title = elem.text.replace('[', '').replace(']', '')

        for elem in child.iter('AbstractText'):
            parts.append(elem.text)
        authors = []
        for author in child.iter('Author'):
            author_elems = []
            for elem in author.iter('ForeName'):
                author_elems.append(elem.text)
            for elem in author.iter('LastName'):
                author_elems.append(elem.text)
            authors.append(" ".join(author_elems))

        author_str = ", ".join(authors)
        abstract = " ".join(parts)

        keyword_list = [keyword.text for keyword in child.iter('Keyword')]
        keywords = ", ".join(keyword_list)

        new_record = {
            "abstract": abstract,
            "title": title,
            "authors": author_str,
            "keywords": keywords,
        }
        records.append(new_record)
    return records


def read_ris(fp):
    """RIS file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the RIS file.
    label: bool
        Check for label. If None, this is automatic.

    Returns
    -------
    list:
        List with entries.

    """

    encodings = ['ISO-8859-1', 'utf-8', 'utf-8-sig']
    entries = None
    for encoding in encodings:
        try:
            with open(fp, 'r', encoding=encoding) as bibliography_file:
                mapping = _tag_key_mapping(reverse=False)
                entries = list(readris(bibliography_file, mapping=mapping))
                break
        except (UnicodeDecodeError, IOError):
            pass

    if entries is None:
        raise ValueError("Cannot find proper encoding for data file.")
    return entries



from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.feature_extraction.text import CountVectorizer

from asreview.feature_extraction.base import BaseFeatureExtraction


class Tfidf(BaseFeatureExtraction):
    """Class to apply SKLearn Tf-idf to texts."""
    name = "tf-idf"

    def __init__(self, *args, ngram_max=1, **kwargs):
        """Initialize tfidf class.

        Arguments
        ---------
        ngram_max: int
            Can use up to ngrams up to ngram_max. For example in the case of
            ngram_max=2, monograms and bigrams could be used.
        """
        super(Tfidf, self).__init__(*args, **kwargs)
        self.ngram_max = ngram_max
        self._model = Pipeline([
            ('vect', CountVectorizer(ngram_range=(1, ngram_max))),
            ('tfidf', TfidfTransformer())]
        )

    def fit(self, texts):
        self._model.fit(texts)

    def transform(self, texts):
        X = self._model.transform(texts).tocsr()
        return X

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_space, hyper_choices = super(Tfidf, self).full_hyper_space()
        hyper_space.update({
            "fex_ngram_max": 1 + hp.randint("fex_ngram_max", 2)
        })
        return hyper_space, hyper_choices



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
from os.path import splitext

from asreview.config import AVAILABLE_CLI_MODI
from asreview.config import AVAILABLE_REVIEW_CLASSES
from asreview.config import DEFAULT_BALANCE_STRATEGY
from asreview.config import DEFAULT_FEATURE_EXTRACTION
from asreview.config import DEFAULT_MODEL
from asreview.config import DEFAULT_N_INSTANCES
from asreview.config import DEFAULT_N_PRIOR_EXCLUDED
from asreview.config import DEFAULT_N_PRIOR_INCLUDED
from asreview.config import DEFAULT_QUERY_STRATEGY
from asreview.config import DEMO_DATASETS
from asreview.config import KERAS_MODELS
from asreview.logging.utils import open_logger
from asreview.readers import ASReviewData
from asreview.review.minimal import MinimalReview
from asreview.review.oracle import ReviewOracle
from asreview.review.simulate import ReviewSimulate
from asreview.settings import ASReviewSettings
from asreview.models.utils import get_model
from asreview.query_strategies.utils import get_query_model
from asreview.balance_strategies.utils import get_balance_model
from asreview.feature_extraction.utils import get_feature_model


def _add_defaults(set_param, default_param):
    set_param.update({key: value for key, value in default_param.items()
                      if key not in set_param})


def get_reviewer(dataset,
                 mode='oracle',
                 model=DEFAULT_MODEL,
                 query_strategy=DEFAULT_QUERY_STRATEGY,
                 balance_strategy=DEFAULT_BALANCE_STRATEGY,
                 feature_extraction=DEFAULT_FEATURE_EXTRACTION,
                 n_instances=DEFAULT_N_INSTANCES,
                 n_papers=None,
                 n_queries=None,
                 embedding_fp=None,
                 verbose=0,
                 prior_included=None,
                 prior_excluded=None,
                 n_prior_included=DEFAULT_N_PRIOR_INCLUDED,
                 n_prior_excluded=DEFAULT_N_PRIOR_EXCLUDED,
                 config_file=None,
                 log_file=None,
                 model_param=None,
                 query_param=None,
                 balance_param=None,
                 feature_param=None,
                 abstract_only=False,
                 extra_dataset=[],
                 **kwargs
                 ):
    """ Get a review object from arguments. See __main__.py for a description
        Of the arguments.
    """

    # Find the URL of the datasets if the dataset is an example dataset.
    if dataset in DEMO_DATASETS.keys():
        dataset = DEMO_DATASETS[dataset]

    cli_settings = ASReviewSettings(
        model=model, n_instances=n_instances, n_queries=n_queries,
        n_papers=n_papers, n_prior_included=n_prior_included,
        n_prior_excluded=n_prior_excluded, query_strategy=query_strategy,
        balance_strategy=balance_strategy,
        feature_extraction=feature_extraction,
        mode=mode, data_fp=dataset,
        abstract_only=abstract_only)
    cli_settings.from_file(config_file)

    if log_file is not None:
        with open_logger(log_file) as logger:
            if logger.is_empty():
                logger.add_settings(cli_settings)
            settings = logger.settings
    else:
        settings = cli_settings
        logger = None

    if n_queries is not None:
        settings.n_queries = n_queries
    if n_papers is not None:
        settings.n_papers = n_papers

    if model_param is not None:
        settings.model_param = model_param
    if query_param is not None:
        settings.query_param = query_param
    if balance_param is not None:
        settings.balance_param = balance_param
    if feature_param is not None:
        settings.feature_param = feature_param

    # Check if mode is valid
    if mode in AVAILABLE_REVIEW_CLASSES:
        logging.info(f"Start review in '{mode}' mode.")
    else:
        raise ValueError(f"Unknown mode '{mode}'.")
    logging.debug(settings)

    as_data = ASReviewData.from_file(dataset, extra_dataset=extra_dataset,
                                     abstract_only=settings.abstract_only)
    texts = as_data.texts
    y = as_data.labels

    data_prior_included, data_prior_excluded = as_data.get_priors()
    if len(data_prior_included) != 0:
        if prior_included is None:
            prior_included = []
        prior_included.extend(data_prior_included.tolist())
    if len(data_prior_excluded) != 0:
        if prior_excluded is None:
            prior_excluded = []
        prior_excluded.extend(data_prior_excluded.tolist())

    if as_data.final_labels is not None:
        with open_logger(log_file) as logger:
            logger.set_final_labels(as_data.final_labels)

    train_model = get_model(settings.model, **settings.model_param)
    query_model = get_query_model(settings.query_strategy,
                                  **settings.query_param)
    balance_model = get_balance_model(settings.balance_strategy,
                                      **settings.balance_param)
    feature_model = get_feature_model(settings.feature_extraction,
                                      **settings.feature_param)

    X = feature_model.fit_transform(texts, as_data.title, as_data.abstract)

    if train_model.name.startswith("lstm-"):
        train_model.embedding_matrix = feature_model.get_embedding_matrix(
            texts, embedding_fp)

    _add_defaults(settings.query_param, query_model.default_param)
    _add_defaults(settings.model_param, train_model.default_param)
    _add_defaults(settings.balance_param, balance_model.default_param)

    if log_file is not None:
        with open_logger(log_file) as logger:
            logger.add_settings(settings)

    # Initialize the review class.
    if mode == "simulate":
        reviewer = ReviewSimulate(
            X, y,
            model=train_model,
            query_model=query_model,
            balance_model=balance_model,
            n_papers=settings.n_papers,
            n_instances=settings.n_instances,
            n_queries=settings.n_queries,
            verbose=verbose,
            prior_included=prior_included,
            prior_excluded=prior_excluded,
            n_prior_included=settings.n_prior_included,
            n_prior_excluded=settings.n_prior_excluded,
            log_file=log_file,
            final_labels=as_data.final_labels,
            **kwargs)
    elif mode == "oracle":
        reviewer = ReviewOracle(
            X,
            model=model,
            query_model=query_model,
            balance_model=balance_model,
            as_data=as_data,
            n_papers=settings.n_papers,
            n_instances=settings.n_instances,
            n_queries=settings.n_queries,
            verbose=verbose,
            prior_included=prior_included,
            prior_excluded=prior_excluded,
            log_file=log_file,
            **kwargs)
    elif mode == "minimal":
        reviewer = MinimalReview(
            X,
            model=model,
            query_model=query_model,
            balance_mode=balance_model,
            n_papers=settings.n_papers,
            n_instances=settings.n_instances,
            n_queries=settings.n_queries,
            verbose=verbose,
            prior_included=prior_included,
            prior_excluded=prior_excluded,
            log_file=log_file,
            **kwargs)
    else:
        raise ValueError("Error finding mode, should never come here...")

    return reviewer


def review(*args, mode="simulate", model=DEFAULT_MODEL, save_model_fp=None,
           **kwargs):
    """ Perform a review from arguments. Compatible with the CLI interface. """
    if mode not in AVAILABLE_CLI_MODI:
        raise ValueError(f"Unknown mode '{mode}'.")

    reviewer = get_reviewer(*args, mode=mode, model=model, **kwargs)

    # Start the review process.
    reviewer.review()

    # If we're dealing with a keras model, we can save the last model weights.
    if save_model_fp is not None and model in KERAS_MODELS:
        save_model_h5_fp = splitext(save_model_fp)[0]+".h5"
        json_model = model.model.to_json()
        with open(save_model_fp, "w") as f:
            json.dump(json_model, f, indent=2)
        model.model.save_weights(save_model_h5_fp, overwrite=True)


def review_oracle(dataset, *args, log_file=None, **kwargs):
    """CLI to the interactive mode."""
    from PyInquirer import prompt, Separator
    if log_file is None:
        while True:
            question = [{
                'type': 'input',
                'name': 'log_file',
                'message': 'Please provide a file to store '
                'the results of your review:'
            }]
            log_file = prompt(question).get("log_file", "")
            if len(log_file) == 0:
                question = [{
                    'type': 'confirm',
                    'message': 'Are you sure you want to continue without'
                    ' saving?',
                    'name': 'force_continue',
                    'default': 'False',
                }]
                force_continue = prompt(question).get('force_continue', False)
                if force_continue:
                    log_file = None
                    break
            else:
                if os.path.isfile(log_file):
                    question = [{
                        'type': 'list',
                        'name': 'action',
                        'message': f'File {log_file} exists, what do you want'
                        ' to do?',
                        'choices': [
                            f'Continue review from {log_file}',
                            f'Delete review in {log_file} and start a new'
                            ' review',
                            f'Choose another file name.',
                            Separator(),
                            f'Exit'
                        ]
                    }]
                    action = prompt(question).get('action', 'Exit')
                    if action == "Exit":
                        return
                    if action.startswith("Continue"):
                        break
                    if action.startswith("Choose another"):
                        continue
                    if action.startswith("Delete"):
                        question = [{
                            'type': 'confirm',
                            'message': f'Are you sure you want to delete '
                            f'{log_file}?',
                            'name': 'delete',
                            'default': 'False',
                        }]
                        delete = prompt(question).get("delete", False)
                        if delete:
                            os.remove(log_file)
                            break
                        else:
                            continue

                break
    try:
        review(dataset, *args, mode='oracle', log_file=log_file, **kwargs)
    except KeyboardInterrupt:
        print('\nClosing down the automated systematic review.')


def review_simulate(dataset, *args, **kwargs):
    """CLI to the oracle mode."""

    review(dataset, *args, mode='simulate', **kwargs)



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
import inspect


class BaseModel(ABC):
    "Abstract class for balance strategies."
    name = "base"

    @property
    def default_param(self):
        """Get the default parameters of the balance strategy.

        Returns
        -------
        dict:
            Dictionary with parameter: default_value
        """
        signature = inspect.signature(self.__init__)
        return {
            k: v.default
            for k, v in signature.parameters.items()
            if v.default is not inspect.Parameter.empty
        }

    def full_hyper_space(self):
        return {}, {}

    def hyper_space(self):
        hyper_space, hyper_choices = self.full_hyper_space()
        return hyper_space, hyper_choices



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.models.base import BaseTrainModel
from sklearn.ensemble import RandomForestClassifier


class RFModel(BaseTrainModel):
    "Random Forest SKLearn model."
    name = "rf"

    def __init__(self, n_estimators=100, max_features=10, class_weight=1.0,
                 random_state=None):
        """Initialize the SKLearn Random Forest model.

        Arguments
        ---------
        n_estimators: int
            Number of estimators.
        max_features: int
            Number of features in the model.
        class_weight: float
            Class weight of the inclusions.
        random_state: int, RandomState
            Set the random state of the RNG.
        """
        super(RFModel, self).__init__()
        if class_weight is not None:
            class_weight = {
                0: 1,
                1: class_weight,
            }

        self._model = RandomForestClassifier(
            n_estimators=int(n_estimators), max_features=int(max_features),
            class_weight=class_weight, random_state=random_state)

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {}
        hyper_space = {
            "mdl_n_estimators": hp.quniform("mdl_n_estimators", 10, 100, 1),
            "mdl_max_features": hp.quniform("mdl_max_features", 6, 10, 1),
            "mdl_class_weight": hp.lognormal('mdl_class_weight', 0, 1),
        }
        return hyper_space, hyper_choices



import sys

import numpy as np

from asreview.feature_extraction.base import BaseFeatureExtraction


class SBERT(BaseFeatureExtraction):
    """Sentence BERT class for feature extraction."""
    name = "sbert"

    def transform(self, texts):
        try:
            from sentence_transformers.SentenceTransformer import SentenceTransformer  #noqa
        except ImportError:
            print("Error: install sentence_transformers package "
                  "(`pip install sentence_transformers`)"
                  " to use Sentence BERT.")
            sys.exit(192)
        model = SentenceTransformer('bert-base-nli-mean-tokens')
        X = np.array(model.encode(texts))
        return X



from pytest import mark
import numpy as np
import os

from asreview.query_strategies.utils import get_query_model
from asreview.models.utils import get_model
from asreview.readers import ASReviewData


@mark.parametrize(
    "query_strategy",
    [
        "max",
        "random",
        "random_max",
        "max_random",
        "uncertainty",
        "max_uncertainty",
        "cluster",
    ])
def test_query(query_strategy, n_features=50, n_sample=100,
               n_instances_list=[0, 1, 5, 50], n_train_idx=[0, 1, 5, 50]):
    classifier = get_model("rf")
    if query_strategy == "cluster":
        data_fp = os.path.join("test", "demo_data", "generic.csv")
        texts = ASReviewData.from_file(data_fp).texts
        while len(texts) < n_features:
            texts = np.append(texts, texts)
            print(len(texts))
#             texts.extend(texts)
        texts = texts[:n_features]
        query_model = get_query_model(
            query_strategy, texts=texts, update_interval=None,
            cluster_size=int(n_sample/3))
    else:
        query_model = get_query_model(query_strategy)
    X = np.random.rand(n_sample, n_features)

    y = np.concatenate((np.zeros(n_sample//2), np.ones(n_sample//2)), axis=0)
    print(X.shape, y.shape)
    order = np.random.permutation(n_sample)
    print(order.shape)
    X = X[order]
    y = y[order]
    sources = query_strategy.split('_')

    classifier.fit(X, y)

    for n_instances in n_instances_list:
        for n_train in n_train_idx:
            shared = {"query_src": {}, "current_queries": {}}
            train_idx = np.random.choice(
                np.arange(n_sample), n_train, replace=False)
            pool_idx = np.delete(np.arange(n_sample), train_idx)
            query_idx, X_query = query_model.query(X, classifier, pool_idx,
                                                   n_instances, shared)
            check_integrity(query_idx, X_query, X, pool_idx, shared,
                            n_instances, sources)


def check_integrity(query_idx, X_query, X, pool_idx, shared,
                    n_instances, sources):
    # First check if the query_indices are valid
    assert len(query_idx) == n_instances
    assert len(query_idx) == len(np.unique(query_idx))
    for i, idx in enumerate(query_idx):
        assert idx in pool_idx
        assert np.all(X_query[i] == X[idx])
        assert idx in shared["current_queries"]

    for idx, src in shared["current_queries"].items():
        assert src in sources
        assert idx in query_idx



from abc import abstractmethod

import numpy as np
from scipy.sparse import issparse, hstack
from asreview.base_model import BaseModel


class BaseFeatureExtraction(BaseModel):
    """Base class for feature extraction methods."""
    name = "base-feature"

    def __init__(self, split_ta=0):
        self.split_ta = split_ta

    def fit_transform(self, texts, titles=None, abstracts=None):
        """Fit and transform a list of texts.

        Arguments
        ---------
        texts: np.array
            A sequence of texts to be transformed. They are not yet tokenized.

        Returns
        -------
        np.array:
            Feature matrix representing the texts.
        """
        self.fit(texts)
        if self.split_ta > 0:
            if titles is None or abstracts is None:
                raise ValueError("Error: if splitting titles and abstracts,"
                                 " supply them!")
            X_titles = self.transform(titles)
            X_abstracts = self.transform(abstracts)
            if issparse(X_titles) and issparse(X_abstracts):
                X_merge = hstack([X_titles, X_abstracts]).tocsr()
            else:
                X_merge = np.concatenate((X_titles, X_abstracts), axis=1)
            return X_merge
        return self.transform(texts)

    def fit(self, texts):
        """Fit the model to the texts.


        It is not always necessary to implement this if there's not real
        fitting being done.

        Arguments
        ---------
        texts: np.array
            Texts to be fitted.
        """
        pass

    @abstractmethod
    def transform(self, texts):
        """Transform a list of texts.

        Arguments
        ---------
        texts: np.array
            A sequence of texts to be transformed. They are not yet tokenized.

        Returns
        -------
        np.array:
            Feature matrix representing the texts.
        """
        raise NotImplementedError

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {}
        hyper_space = {
            "fex_split_ta": hp.randint("fex_split_ta", 2),
        }
        return hyper_space, hyper_choices



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from sklearn.naive_bayes import MultinomialNB

from asreview.models.base import BaseTrainModel


class NBModel(BaseTrainModel):
    "Naive Bayes SKLearn model."
    name = "nb"

    def __init__(self, alpha=3.822):
        """Initialize the SKLearn Naive Bayes model.

        Arguments:
        ----------
        alpha: float
            Parameter to set the regularization strength of the model.
        """
        super(NBModel, self).__init__()
        self._model = MultinomialNB(alpha=alpha)
        logging.debug(self._model)

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {}
        hyper_space = {
            "mdl_alpha": hp.lognormal("mdl_alpha", 0, 1),
        }
        return hyper_space, hyper_choices



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def is_pickle(fp):
    """Check file for pickle extension."""
    fp = str(fp)
    return fp.endswith('.pkl') or fp.endswith('.pickle')


def convert_list_type(x, to_type=int):
    """Convert elements in list to given type."""

    return list(map(to_type, x))



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from contextlib import contextmanager


def _get_logger_class(fp):
    "Get logging class from file extension."
    from asreview.logging.hdf5 import HDF5Logger
    from asreview.logging.json import JSONLogger
    from asreview.logging.dict import DictLogger

    if fp is None:
        return DictLogger

    log_ext = os.path.splitext(fp)[1]
    if log_ext in ['.h5', '.hdf5', '.he5']:
        logger_class = HDF5Logger
    else:
        logger_class = JSONLogger
    return logger_class


@contextmanager
def open_logger(fp, *args, read_only=False, **kwargs):
    """Open a logger from a file.

    Arguments
    ---------
    fp: str
        File to open.
    read_only: bool
        Whether to open the file in read_only mode.

    Returns
    -------
    BaseLogger:
        Depending on the extension the appropriate logger is
        chosen:
        - [.h5, .hdf5, .he5] -> HDF5Logger.
        - None -> DictLogger (doesn't store anything permanently).
        - Anything else -> JSONLogger.
    """
    logger_class = _get_logger_class(fp)
    try:
        logger = logger_class(log_fp=fp, *args, read_only=read_only, **kwargs)
        yield logger
    finally:
        logger.close()


def loggers_from_dir(data_dir, prefix="result"):
    """Obtain a list of loggers from a directory.

    Arguments
    ---------
    data_dir: str
        Directory where to search for logging files.
    prefix: str
        Files starting with the prefix are assumed to be logging files.
        The rest is ignored.

    Returns
    -------
    dict:
        A dictionary of opened loggers, with their (base) filenames as keys.
    """
    loggers = {}
    files = os.listdir(data_dir)
    if not files:
        print(f"Error: {data_dir} is empty")
        return None

    for log_file in files:
        if not log_file.startswith(prefix):
            continue

        log_fp = os.path.join(data_dir, log_file)
        logger_class = _get_logger_class(log_fp)
        loggers[log_file] = logger_class(log_fp=log_fp, read_only=True)

    return loggers



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import numpy as np

from asreview.feature_extraction.base import BaseFeatureExtraction


def _train_model(corpus, *args, **kwargs):
    import gensim

    model = gensim.models.doc2vec.Doc2Vec(*args, **kwargs)
    model.build_vocab(corpus)
    model.train(corpus, total_examples=model.corpus_count,
                epochs=model.epochs)
    return model


def _transform_text(model, corpus):
    X = []
    for doc_id in range(len(corpus)):
        doc_vec = model.infer_vector(corpus[doc_id].words)
        X.append(doc_vec)
    return np.array(X)


class Doc2Vec(BaseFeatureExtraction):
    """Base class for doc2vec feature extraction."""
    name = "doc2vec"

    def __init__(self, *args, vector_size=40, epochs=33, min_count=1,
                 workers=1, window=7, dm_concat=0, dm=2, dbow_words=0,
                 **kwargs):
        """Initialize the doc2vec model.

        Arguments
        ---------
        vector_size: int
            Output size of the vector.
        epochs: int
            Number of epochs to train the doc2vec model.
        min_count: int
            Minimum number of occurences for a word in the corpus for it to
            be included in the model.
        workers: int
            Number of threads to train the model with.
        window: int
            Maximum distance over which word vectors influence each other.
        dm_concat: int
            Whether to concatenate word vectors or not.
            See paper for more detail.
        dm: int
            Model to use.
            0: Use distribute bag of words (DBOW).
            1: Use distributed memory (DM).
            2: Use both of the above with half the vector size and concatenate
            them.
        dbow_words: int
            Whether to train the word vectors using the skipgram method.
        """
        super(Doc2Vec, self).__init__(*args, **kwargs)
        self.vector_size = int(vector_size)
        self.epochs = int(epochs)
        self.min_count = int(min_count)
        self.workers = int(workers)
        self.window = int(window)
        self.dm_concat = int(dm_concat)
        self.dm = int(dm)
        self.dbow_words = int(dbow_words)
        self._model = None
        self._model_dm = None
        self._model_dbow = None

    def fit(self, texts):
        try:
            from gensim.utils import simple_preprocess
            from gensim.models.doc2vec import TaggedDocument
        except ImportError:
            print("Error: install gensim package (`pip install gensim`) to use"
                  " doc2vec.")
            sys.exit(192)

        model_param = {
            "vector_size": self.vector_size,
            "epochs": self.epochs,
            "min_count": self.min_count,
            "workers": self.workers,
            "window": self.window,
            "dm_concat": self.dm_concat,
            "dbow_words": self.dbow_words,
        }

        corpus = [TaggedDocument(simple_preprocess(text), [i])
                  for i, text in enumerate(texts)]

        if self.dm == 2:
            model_param["vector_size"] = int(model_param["vector_size"]/2)
            self.model_dm = _train_model(corpus, **model_param, dm=1)
            self.model_dbow = _train_model(corpus, **model_param, dm=0)
        else:
            self.model = _train_model(corpus, **model_param, dm=self.dm)

    def transform(self, texts):
        try:
            from gensim.utils import simple_preprocess
            from gensim.models.doc2vec import TaggedDocument
        except ImportError:
            print("Error: install gensim package (`pip install gensim`) to use"
                  " doc2vec.")
            sys.exit(192)
        corpus = [TaggedDocument(simple_preprocess(text), [i])
                  for i, text in enumerate(texts)]

        if self.dm == 2:
            X_dm = _transform_text(self.model_dm, corpus)
            X_dbow = _transform_text(self.model_dbow, corpus)
            X = np.concatenate((X_dm, X_dbow), axis=1)
        else:
            X = _transform_text(self.model, corpus)
        return X

    def full_hyper_space(self):
        from hyperopt import hp
        eps = 1e-7

        hyper_space, hyper_choices = super(Doc2Vec, self).full_hyper_space()
        hyper_space.update({
            "fex_vector_size": hp.quniform(
                "fex_vector_size", 31.5, 127.5-eps, 8),
            "fex_epochs": hp.quniform("fex_epochs", 20, 50, 1),
            "fex_min_count": hp.quniform("fex_min_count", 0.5, 2.499999, 1),
            "fex_window": hp.quniform("fex_window", 4.5, 9.4999999, 1),
            "fex_dm_concat": hp.randint("fex_dm_concat", 2),
            "fex_dm": hp.randint("fex_dm", 3),
            "fex_dbow_words": hp.randint("fex_dbow_words", 2),
        })

        return hyper_space, hyper_choices



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json
import os

import numpy as np
from scipy import stats
from sklearn.cluster import KMeans

from asreview.logging.utils import loggers_from_dir
from asreview.analysis.statistics import _get_labeled_order
from asreview.analysis.statistics import _get_limits
from asreview.analysis.statistics import _find_inclusions
from asreview.analysis.statistics import _get_last_proba_order
from asreview.analysis.statistics import _random_ttd, _cluster_ttd, _max_ttd


class Analysis():
    """ Analysis object to do statistical analysis on log files. """
    def __init__(self, loggers, key=None):
        """
        Arguments
        ---------
        loggers: list, BaseLogger
            Either a list of loggers (opened files) or a single logger.
        key: str
            Give a name to the analysis.
        """
        if isinstance(loggers, list):
            loggers = {i: logger for i, logger in enumerate(loggers)}

        # Sometimes an extra dataset is present in the log_file(s).
        # These signify not the labels on which the model was trained, but the
        # ones that were included in the end (or some other intermediate step.
        self.final_labels = None
        self.labels = None
        self.empty = True

        if loggers is None:
            return

        self.key = key
        self.loggers = loggers
        self.num_runs = len(self.loggers)
        if self.num_runs == 0:
            return

        self._first_file = list(self.loggers.keys())[0]
        self.labels = self.loggers[self._first_file].get('labels')
        try:
            self.final_labels = self.loggers[self._first_file].get(
                'final_labels')
        except KeyError:
            pass
        self.empty = False
        self.inc_found = {}

    @classmethod
    def from_dir(cls, data_dir):
        """Create an Analysis object from a directory."""
        key = os.path.basename(os.path.normpath(data_dir))
        loggers = loggers_from_dir(data_dir)
        analysis_inst = Analysis(loggers, key=key)
        if analysis_inst.empty:
            return None

        analysis_inst.data_dir = data_dir
        return analysis_inst

    def inclusions_found(self, result_format="fraction", final_labels=False,
                         **kwargs):
        """Get the number of inclusions at each point in time.

        Caching is used to prevent multiple calls being expensive.

        Arguments
        ---------
        result_format: str
            The format % or # of the returned values.
        final_labels: bool
            If true, use the final_labels instead of labels for analysis.

        Returns
        -------
        tuple:
            Three numpy arrays with x, y, error_bar.
        """
        if final_labels:
            labels = self.final_labels
        else:
            labels = self.labels

        fl = final_labels
        if fl not in self.inc_found:
            # Compute the comclusions if not found in cache.
            self.inc_found[fl] = {}
            avg, err, iai, ninit = self._get_inc_found(**kwargs, labels=labels)
            self.inc_found[fl]["avg"] = avg
            self.inc_found[fl]["err"] = err
            self.inc_found[fl]["inc_after_init"] = iai
            self.inc_found[fl]["n_initial"] = ninit
        dx = 0
        dy = 0
        x_norm = len(labels)-self.inc_found[fl]["n_initial"]
        y_norm = self.inc_found[fl]["inc_after_init"]

        if result_format == "percentage":
            x_norm /= 100
            y_norm /= 100
        elif result_format == "number":
            x_norm /= len(labels)
            y_norm /= self.inc_found[fl]["inc_after_init"]

        norm_xr = (np.arange(len(self.inc_found[fl]["avg"]))-dx)/x_norm
        norm_yr = (np.array(self.inc_found[fl]["avg"])-dy)/y_norm
        norm_y_err = np.array(self.inc_found[fl]["err"])/y_norm

        return norm_xr, norm_yr, norm_y_err

    def _get_inc_found(self, labels=False):
        """Get the number of inclusions (without formatting)."""
        inclusions_found = []

        for logger in self.loggers.values():
            inclusions, inc_after_init, n_initial = _find_inclusions(
                logger, labels)
            inclusions_found.append(inclusions)

        inc_found_avg = []
        inc_found_err = []
        for i_instance in itertools.count():
            cur_vals = []
            for i_file in range(self.num_runs):
                try:
                    cur_vals.append(inclusions_found[i_file][i_instance])
                except IndexError:
                    pass
            if len(cur_vals) == 0:
                break
            if len(cur_vals) == 1:
                err = cur_vals[0]
            else:
                err = stats.sem(cur_vals)
            avg = np.mean(cur_vals)
            inc_found_avg.append(avg)
            inc_found_err.append(err)

        if self.num_runs == 1:
            inc_found_err = np.zeros(len(inc_found_err))

        return inc_found_avg, inc_found_err, inc_after_init, n_initial

    def wss(self, val=100, x_format="percentage", **kwargs):
        """Get the WSS (Work Saved Sampled) value.

        Arguments
        ---------
        val:
            At which recall, between 0 and 100.
        x_format:
            Format for position of WSS value in graph.

        Returns
        -------
        tuple:
            Tuple consisting of WSS value, x_positions, y_positions of WSS bar.
        """
        norm_xr, norm_yr, _ = self.inclusions_found(
            result_format="percentage", **kwargs)

        if x_format == "number":
            x_return, y_result, _ = self.inclusions_found(
                result_format="number", **kwargs)
            y_max = self.inc_found[False]["inc_after_init"]
            y_coef = y_max/(len(self.labels) -
                            self.inc_found[False]["n_initial"])
        else:
            x_return = norm_xr
            y_result = norm_yr
            y_max = 1.0
            y_coef = 1.0

        for i in range(len(norm_yr)):
            if norm_yr[i] >= val - 1e-6:
                return (norm_yr[i] - norm_xr[i],
                        (x_return[i], x_return[i]),
                        (x_return[i]*y_coef, y_result[i]))
        return (None, None, None)

    def rrf(self, val=10, x_format="percentage", **kwargs):
        """Get the RRF (Relevant References Found).

        Arguments
        ---------
        val:
            At which recall, between 0 and 100.
        x_format:
            Format for position of RRF value in graph.

        Returns
        -------
        tuple:
            Tuple consisting of RRF value, x_positions, y_positions of RRF bar.

        """
        norm_xr, norm_yr, _ = self.inclusions_found(
            result_format="percentage", **kwargs)

        if x_format == "number":
            x_return, y_return, _ = self.inclusions_found(
                result_format="number", **kwargs)
        else:
            x_return = norm_xr
            y_return = norm_yr

        for i in range(len(norm_yr)):
            if norm_xr[i] >= val - 1e-6:
                return (norm_yr[i],
                        (x_return[i], x_return[i]),
                        (0, y_return[i]))
        return (None, None, None)

    def avg_time_to_discovery(self):
        """Get the best/last estimate on how long it takes to find a paper.

        Returns
        -------
        dict:
            For each inclusion, key=paper_id, value=avg time.
        """
        labels = self.labels

        one_labels = np.where(labels == 1)[0]
        time_results = {label: [] for label in one_labels}
        n_initial = []

        for i_file, logger in enumerate(self.loggers.values()):
            label_order, n = _get_labeled_order(logger)
            proba_order = _get_last_proba_order(logger)
            n_initial.append(n)

            for i_time, idx in enumerate(label_order):
                if labels[idx] == 1:
                    time_results[idx].append(i_time)

            for i_time, idx in enumerate(proba_order):
                if labels[idx] == 1 and len(time_results[idx]) <= i_file:
                    time_results[idx].append(i_time + len(label_order))

            for idx in time_results:
                if len(time_results[idx]) <= i_file:
                    time_results[idx].append(
                        len(label_order) + len(proba_order))

        results = {}
        for label in time_results:
            trained_time = []
            for i_file, time in enumerate(time_results[label]):
                if time >= n_initial[i_file]:
                    trained_time.append(time)
            if len(trained_time) == 0:
                results[label] = 0
            else:
                results[label] = np.average(trained_time)
        return results

    def limits(self, prob_allow_miss=[0.1]):
        """For each query, compute the number of papers for a criterium.

        A criterium is the average number of papers missed. For example,
        with 0.1, the criterium is that after reading x papers, there is
        (about) a 10% chance that one paper is not included. Another example,
        with 2.0, there are on average 2 papers missed after reading x papers.
        The value for x is returned for each query and probability by the
        function.

        Arguments
        ---------
        prob_allow_miss: list, float
            Sets the criterium for how many papers can be missed.

        returns
        -------
        dict:
            One entry, "x_range" with the number of papers read.
            List, "limits" of results for each probability and
            at # papers read.
        """
        if not isinstance(prob_allow_miss, list):
            prob_allow_miss = [prob_allow_miss]
        logger = self.loggers[self._first_file]
        n_queries = logger.n_queries()
        results = {
            "x_range": [],
            "limits": [[] for _ in range(len(prob_allow_miss))],
        }

        n_train = 0
        for query_i in range(n_queries):
            new_limits = _get_limits(self.loggers, query_i, self.labels,
                                     proba_allow_miss=prob_allow_miss)

            try:
                new_train_idx = logger.get("train_idx", query_i)
            except KeyError:
                new_train_idx = None

            if new_train_idx is not None:
                n_train = len(new_train_idx)

            if new_limits is not None:
                results["x_range"].append(n_train)
                for i_prob in range(len(prob_allow_miss)):
                    results["limits"][i_prob].append(new_limits[i_prob])

        results["x_range"] = np.array(results["x_range"], dtype=np.int)
        for i_prob in range(len(prob_allow_miss)):
            results["limits"][i_prob] = np.array(
                results["limits"][i_prob], np.int)
        return results

    def time_to_inclusion(self, X_fp):
        with open(X_fp, "r") as fp:
            X = np.array(json.load(fp))

        n_clusters = int(len(self.labels)/150)
        model = KMeans(n_clusters=n_clusters, n_init=1)
        clusters = model.fit_predict(X, self.labels)
        logger = self.loggers[self._first_file]
        n_queries = logger.n_queries()

        results = {
            "x_range": [],
            "ttd": {
                "random": [],
                "cluster": [],
                "max": [],
            }
        }
        n_train = 0
        for query_i in range(n_queries):
            new_random_ttd = _random_ttd(self.loggers, query_i, self.labels)
            new_cluster_ttd = _cluster_ttd(self.loggers, query_i, self.labels,
                                           clusters)
            new_max_ttd = _max_ttd(self.loggers, query_i, self.labels)

            results["ttd"]["random"].append(new_random_ttd)
            results["ttd"]["cluster"].append(new_cluster_ttd)
            results["ttd"]["max"].append(new_max_ttd)

            try:
                new_train_idx = logger.get("train_idx", query_i)
            except KeyError:
                new_train_idx = None

            if new_train_idx is not None:
                n_train = len(new_train_idx)
            results["x_range"].append(n_train)

        return results

    def close(self):
        "Close loggers."
        for logger in self.loggers.values():
            logger.close()



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from datetime import datetime

import numpy as np

from asreview.analysis.statistics import _get_labeled_order
from asreview.analysis.statistics import _get_last_proba_order
from asreview.logging.base import BaseLogger


def get_serial_list(array, dtype=None):
    if isinstance(array, np.ndarray):
        array = array.tolist()
    if dtype is not None:
        array = list(map(dtype, array))
    return array


class DictLogger(BaseLogger):
    """Class for logging the Systematic Review with no storage."""
    version = "1.1"
    read_only = False

    def __init__(self, log_fp, *_, **__):
        super(DictLogger, self).__init__(log_fp)
        self.read_only = False

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.close()

    def __str__(self):
        return self._print_logs()

    def _print_logs(self):
        self._log_dict["time"]["end_time"] = str(datetime.now())
        label_order, _ = _get_labeled_order(self)
        try:
            labels_assigned = self.get("labels")[label_order]
        except (KeyError, IndexError):
            return ""
        labels = list(zip(label_order, labels_assigned))

        log_str = "Labeled during review:\n\n"
        for label in labels:
            log_str += f"{label[0]} => {label[1]}\n"

        pool_order = _get_last_proba_order(self)
        if len(pool_order) > 0:
            log_str += "\n\n Most likely included according to ASReview:\n\n"
            for idx in pool_order:
                log_str += f"{idx}\n"

        return log_str

    def _add_to_log(self, new_dict, i, append_result=False):
        # Find the first number that is not logged yet.
        results = self._log_dict["results"]
        if i is None:
            if (len(results) > 0
                    and set(new_dict.keys()).isdisjoint(results[-1])):
                i = len(results)-1
            else:
                i = len(results)

        while i >= len(results):
            results.append({})

        for key in new_dict:
            if key in results[i] and append_result:
                results[i][key].extend(new_dict[key])
            else:
                results[i][key] = new_dict[key]

    def is_empty(self):
        return len(self._log_dict["results"]) == 0

    def set_labels(self, y):
        self._log_dict["labels"] = y.tolist()

    def set_final_labels(self, y):
        self._log_dict["final_labels"] = y.tolist()

    def add_settings(self, settings):
        self.settings = settings
        self._log_dict["settings"] = vars(settings)

    def add_classification(self, idx, labels, methods, query_i):
        # Ensure that variables are serializable
        idx = get_serial_list(idx, int)
        labels = get_serial_list(labels, int)
        methods = get_serial_list(methods, str)

        new_dict = {'labelled': list(zip(idx, labels, methods))}
        self._add_to_log(new_dict, query_i, append_result=True)

    def add_proba(self, pool_idx, train_idx, proba, query_i):
        new_dict = {
            "pool_idx": get_serial_list(pool_idx, int),
            "train_idx": get_serial_list(train_idx, int),
            "proba": get_serial_list(proba, float),
        }
        self._add_to_log(new_dict, query_i)

    def n_queries(self):
        return len(self._log_dict["results"])

    def get(self, variable, query_i=None, idx=None):
        if query_i is not None:
            res = self._log_dict["results"][query_i]

        array = None
        label_vars = ["label_idx", "inclusions", "label_methods"]
        if variable in label_vars:
            array_id = label_vars.index(variable)
            if variable == "label_methods":
                dtype = str
            else:
                dtype = np.int
            array = np.array([x[array_id] for x in res["labelled"]],
                             dtype=dtype)
        elif variable == "labels":
            array = np.array(self._log_dict["labels"], dtype=np.int)
        elif variable == "final_labels":
            array = np.array(self._log_dict["final_labels"], dtype=np.int)
        elif variable == "proba":
            array = np.array(res["proba"], dtype=np.float)
        elif variable == "train_idx":
            array = np.array(res["train_idx"], dtype=np.int)
        elif variable == "pool_idx":
            array = np.array(res["pool_idx"], dtype=np.int)
        if array is None:
            return None
        if idx is not None:
            return array[idx]
        return array

    def delete_last_query(self):
        self._log_dict["results"].pop()

    def initialize_structure(self):
        from asreview import __version__ as asr_version
        self._log_dict = OrderedDict({
            "time": {"start_time": str(datetime.now())},
            "version": self.version,
            "software_version": asr_version,
            "settings": {},
            "results": [],
        })

    def close(self):
        if not self.read_only:
            self.save()

    def save(self):
        print(self)

    def restore(self, *_, **__):
        self.initialize_structure()



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Uncertainty sampling while saving probabilities.
'''

# from typing import Tuple

import numpy as np
from modAL.utils.selection import multi_argmax
from modAL.utils.selection import shuffled_argmax


from asreview.query_strategies.base import ProbaQueryStrategy


class UncertaintyQuery(ProbaQueryStrategy):
    "Maximum uncertainty query strategy."
    name = "uncertainty"

    def __init__(self, random_tie_break=False):
        """Initialize the maximum uncertainty query strategy.

        Arguments:
        ----------
        random_tie_break: bool
            If true randomly decide which ones to include by tie-break.
        """
        super(UncertaintyQuery, self).__init__()
        self.random_tie_break = random_tie_break

    def _query(self, X, pool_idx, n_instances=1, proba=None):
        uncertainty = 1 - np.max(proba[pool_idx], axis=1)
        if not self.random_tie_break:
            query_idx = multi_argmax(uncertainty, n_instances=n_instances)
        else:
            query_idx = shuffled_argmax(uncertainty, n_instances=n_instances)

        return pool_idx[query_idx], X[pool_idx[query_idx]]



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# based on https://github.com/pypa/sampleproject - MIT License

# Always prefer setuptools over distutils
import re
from setuptools import setup, find_packages
from os import path
from io import open

import versioneer


def get_long_description():
    """Get project description based on README"""
    here = path.abspath(path.dirname(__file__))

    # Get the long description from the README file
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()

    # remove emoji
    long_description = re.sub(r"\:[a-z_]+\:", "", long_description)

    return long_description


DEPS = {
    "sbert": ['sentence_transformers'],
    "doc2vec": ['gensim'],
    "dev": ['check-manifest'],
    'test': ['coverage', 'pytest'],
    'performance': ['python-Levenshtein'],
}
DEPS['all'] = DEPS['sbert'] + DEPS['doc2vec'] + DEPS['dev']
DEPS['all'] += DEPS['performance']

setup(
    name='asreview',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Automated Systematic Review',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/msdslab/automated-systematic-review',
    author='ASReview Core Development Team',
    author_email='asreview@uu.nl',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='systematic review',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    python_requires='~=3.6',
    install_requires=[
        'numpy',
        'tensorflow',
        'sklearn',
        'pandas',
        'modAL',
        'RISparser',
        'dill',
        'PyInquirer',
        'fuzzywuzzy',
        'h5py',
        'xlrd>=1.0.0',
    ],
    extras_require=DEPS,
    entry_points={
        'console_scripts': [
            'asreview=asreview.__main__:main',
            'asr=asreview.__main__:main_depr',
        ],
    },
    project_urls={
        'Bug Reports':
            'https://github.com/msdslab/automated-systematic-review/issues',
        'Source': 'https://github.com/msdslab/automated-systematic-review/',
    },
)



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Initial sample.

This module is used to draw an initial sample. This sample is the
so-called 'pre-knowledge' of the researcher.
"""
import numpy as np


def sample_prior_knowledge(
        labels, n_prior_included=10,
        n_prior_excluded=10, random_state=None):
    """Function to sample prelabelled articles.

    Arguments
    ---------
    labels: np.ndarray
        Labels in a 2d numpy array (the result of
        keras.utils.to_categorical).
    n_included: int
        The number of positive labels.
    n_excluded: int
        The number of negative labels.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    np.ndarray:
        An array with n_included and n_excluded indices.

    """

    # set random state
    r = np.random.RandomState(random_state)

    # retrieve the index of included and excluded papers
    included_indexes = np.where(labels == 1)[0]
    excluded_indexes = np.where(labels == 0)[0]

    if len(included_indexes) < n_prior_included:
        print(f"Found only {len(included_indexes)}, "
              f"when I need {n_prior_included}.")
    # select randomly from included and excluded papers
    included_indexes_sample = r.choice(
        included_indexes, n_prior_included, replace=False)
    excluded_indexes_sample = r.choice(
        excluded_indexes, n_prior_excluded, replace=False)

    init = np.append(included_indexes_sample, excluded_indexes_sample)

    return init



import os
from shutil import copyfile

import numpy as np

from asreview.logging import open_logger
from asreview.review.factory import get_reviewer

data_fp = os.path.join("test", "demo_data", "generic_labels.csv")
embedding_fp = os.path.join("test", "demo_data", "generic.vec")
cfg_dir = os.path.join("test", "cfg_files")
log_dir = os.path.join("test", "log_files")
h5_log_file = os.path.join(log_dir, "test.h5")
json_log_file = os.path.join(log_dir, "test.json")


def test_log_continue_json():
    inter_file = os.path.join(log_dir, "test_1_inst.json")
    if not os.path.isfile(inter_file):
        reviewer = get_reviewer(
            data_fp, mode="simulate", model="nb", embedding_fp=embedding_fp,
            prior_included=[1, 3], prior_excluded=[2, 4], log_file=inter_file,
            n_instances=1, n_queries=1)
        reviewer.review()

    copyfile(inter_file, json_log_file)
    check_model(mode="simulate", model="nb", log_file=json_log_file,
                continue_from_log=True, n_instances=1, n_queries=2)


def test_log_continue_h5():
    inter_file = os.path.join(log_dir, "test_1_inst.h5")
    if not os.path.isfile(inter_file):
        reviewer = get_reviewer(
            data_fp, mode="simulate", model="nb", embedding_fp=embedding_fp,
            prior_included=[1, 3], prior_excluded=[2, 4], log_file=inter_file,
            n_instances=1, n_queries=1)
        reviewer.review()
    copyfile(inter_file, h5_log_file)
    check_model(mode="simulate", model="nb", log_file=h5_log_file,
                continue_from_log=True, n_instances=1, n_queries=2)


def test_lstm_base():
    check_model(mode="simulate",
                config_file=os.path.join(cfg_dir, "lstm_base.ini"),
                log_file=h5_log_file)


def test_lstm_pool():
    check_model(mode="simulate",
                config_file=os.path.join(cfg_dir, "lstm_pool.ini"),
                log_file=json_log_file)


def test_nb():
    check_model(mode="simulate",
                model="nb",
                log_file=None,
                use_granular=True,
                n_instances=1, n_queries=1)


def test_svm():
    check_model(mode="simulate",
                model="svm",
                log_file=json_log_file,
                use_granular=False,
                n_instances=1, n_queries=2)


def test_rf():
    check_model(mode="simulate",
                model="rf",
                log_file=json_log_file,
                use_granular=False,
                n_instances=1, n_queries=2)


def test_nn_2_layer():
    check_model(mode="simulate",
                model="nn-2-layer",
                log_file=json_log_file,
                n_instances=1, n_queries=2)


def test_logistic():
    check_model(mode="simulate",
                model="logistic",
                log_file=json_log_file,
                n_instances=1, n_queries=2)


def check_label_methods(label_methods, n_labels, methods):
    assert len(label_methods) == n_labels
    for method in label_methods:
        assert method in methods


def check_log(logger):

    check_label_methods(logger.get("label_methods", 0), 4, ["initial"])
    check_label_methods(logger.get("label_methods", 1), 1, ["max", "random"])
    check_label_methods(logger.get("label_methods", 2), 1, ["max", "random"])

    assert len(logger.get("inclusions", 0)) == 4
    assert len(logger.get("inclusions", 1)) == 1
    assert len(logger.get("inclusions", 2)) == 1

    assert len(logger.get("train_idx", 1)) == 4
    assert len(logger.get("pool_idx", 1)) == 2

    assert len(logger.get("train_idx", 2)) == 5
    assert len(logger.get("pool_idx", 2)) == 1

    assert len(logger.get("labels")) == 6


def check_model(monkeypatch=None, use_granular=False, log_file=h5_log_file,
                continue_from_log=False, mode="oracle", **kwargs):
    if not continue_from_log:
        try:
            if log_file is not None:
                os.unlink(log_file)
        except OSError:
            pass

    if monkeypatch is not None:
        monkeypatch.setattr('builtins.input', lambda _: "0")
    # start the review process.
    reviewer = get_reviewer(data_fp, mode=mode, embedding_fp=embedding_fp,
                            prior_included=[1, 3], prior_excluded=[2, 4],
                            log_file=log_file,
                            **kwargs)
    if use_granular:
        with open_logger(log_file) as logger:
            # Two loops of training and classification.
            reviewer.train()
            reviewer.log_probabilities(logger)
            query_idx = reviewer.query(1)
            inclusions = reviewer._get_labels(query_idx)
            reviewer.classify(query_idx, inclusions, logger)

            reviewer.train()
            reviewer.log_probabilities(logger)
            query_idx = reviewer.query(1)
            inclusions = reviewer._get_labels(query_idx)
            reviewer.classify(query_idx, inclusions, logger)
    else:
        with open_logger(log_file) as logger:
            if log_file is None:
                logger.set_labels(reviewer.y)
                init_idx, init_labels = reviewer._prior_knowledge()
                reviewer.query_i = 0
                reviewer.train_idx = np.array([], dtype=np.int)
                reviewer.classify(init_idx, init_labels, logger,
                                  method="initial")

            reviewer._do_review(logger)
            if log_file is None:
                print(logger._log_dict)
                check_log(logger)

    if log_file is not None:
        with open_logger(log_file, read_only=True) as logger:
            check_log(logger)
